from ..bundled.traitlets import HasTraits, Instance
from ..errors import OkonomiyakiError

from .epd_platform import EPDPlatform

# This data representation is a bit stupid, but we keep it as is from
# buildware/epd_repo to make it easier to sync.
_SUBDIR = [
    # short     subdir                 arch     platform  osdist
    ('win-64', 'Windows/amd64', 'amd64', 'win32', None),
    ('win-32', 'Windows/x86', 'x86', 'win32', None),
    ('osx-64', 'MacOSX/amd64', 'amd64', 'darwin', None),
    ('osx-32', 'MacOSX/x86', 'x86', 'darwin', None),
    ('rh3-64', 'RedHat/RH3_amd64', 'amd64', 'linux2', 'RedHat_3'),
    ('rh3-32', 'RedHat/RH3_x86', 'x86', 'linux2', 'RedHat_3'),
    ('rh5-64', 'RedHat/RH5_amd64', 'amd64', 'linux2', 'RedHat_5'),
    ('rh5-32', 'RedHat/RH5_x86', 'x86', 'linux2', 'RedHat_5'),
    ('rh6-64', 'RedHat/RH6_amd64', 'amd64', 'linux2', 'RedHat_6'),
    ('rh7-64', 'RedHat/RH7_amd64', 'amd64', 'linux2', 'RedHat_7'),
    ('sol-64', 'Solaris/Sol10_amd64', 'amd64', 'sunos5', 'Solaris_10'),
    ('sol-32', 'Solaris/Sol10_x86', 'x86', 'sunos5', 'Solaris_10'),
]


class LegacyEPDPlatform(HasTraits):
    _epd_platform = Instance(EPDPlatform)

    @classmethod
    def from_running_system(cls, arch=None):
        return cls(_epd_platform=EPDPlatform.from_running_system(arch))

    @classmethod
    def from_arch_and_platform(cls, arch, platform):
        epd_platform_string = _get_entry_from_arch_platform(arch, platform)[0]
        return cls.from_epd_platform_string(epd_platform_string)

    @classmethod
    def from_arch_and_osdist(cls, arch, osdist):
        epd_platform_string = _get_entry_from_arch_osdist(arch, osdist)[0]
        return cls.from_epd_platform_string(epd_platform_string)

    @classmethod
    def from_epd_platform_string(cls, epd_platform_string):
        _epd_platform = EPDPlatform.from_epd_string(epd_platform_string)
        return cls(_epd_platform)

    def __init__(self, _epd_platform, **kw):
        super(LegacyEPDPlatform, self).__init__(_epd_platform=_epd_platform,
                                                **kw)

    @property
    def arch(self):
        return self._epd_platform.arch

    @property
    def osdist(self):
        entry = _get_entry(self.short)
        return entry[4]

    @property
    def platform(self):
        entry = _get_entry(self.short)
        return entry[3]

    @property
    def short(self):
        return self._epd_platform.short

    @property
    def subdir(self):
        entry = _get_entry(self.short)
        return entry[1]

    @property
    def _comp_key(self):
        return (self.arch, self.platform, self.osdist)

    def __eq__(self, other):
        if not isinstance(other, LegacyEPDPlatform):
            return NotImplemented
        return self._comp_key == other._comp_key

    def __ne__(self, other):
        if not isinstance(other, LegacyEPDPlatform):
            return NotImplemented
        return self._comp_key != other._comp_key

    def __str__(self):
        return self.short


def _get_entry(short):
    for entry in _SUBDIR:
        if entry[0] == short:
            return entry
    raise OkonomiyakiError("Invalid short epd string: {0}".format(short))


def _get_entry_from_arch_osdist(arch, osdist):
    for entry in _SUBDIR:
        if entry[2] == arch and entry[-1] == osdist:
            return entry
    raise OkonomiyakiError("Invalid arch/osdist combination: {0}/{1}".
                           format(arch, osdist))


def _get_entry_from_arch_platform(arch, platform):
    for entry in _SUBDIR:
        if entry[2] == arch and entry[-2] == platform:
            return entry
    raise OkonomiyakiError("Invalid arch/platform combination: {0}/{1}".
                           format(arch, platform))



# encoding: utf-8
"""
A lightweight Traits like module.

This is designed to provide a lightweight, simple, pure Python version of
many of the capabilities of enthought.traits.  This includes:

* Validation
* Type specification with defaults
* Static and dynamic notification
* Basic predefined types
* An API that is similar to enthought.traits

We don't support:

* Delegation
* Automatic GUI generation
* A full set of trait types.  Most importantly, we don't provide container
  traits (list, dict, tuple) that can trigger notifications if their
  contents change.
* API compatibility with enthought.traits

There are also some important difference in our design:

* enthought.traits does not validate default values.  We do.

We choose to create this module because we need these capabilities, but
we need them to be pure Python so they work in all Python implementations,
including Jython and IronPython.

Inheritance diagram:

.. inheritance-diagram:: IPython.utils.traitlets
   :parts: 3

Authors:

* Brian Granger
* Enthought, Inc.  Some of the code in this file comes from enthought.traits
  and is licensed under the BSD license.  Also, many of the ideas also come
  from enthought.traits even though our implementation is very different.
"""

#-----------------------------------------------------------------------------
#  Copyright (C) 2008-2011  The IPython Development Team
#
#  Distributed under the terms of the BSD License.  The full license is in
#  the file COPYING, distributed as part of this software.
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------


import inspect
import re
import sys
import types
from types import FunctionType
try:
    from types import ClassType, InstanceType
    ClassTypes = (ClassType, type)
except:
    ClassTypes = (type,)

from ...utils import py3compat


def import_item(name):
    """Import and return ``bar`` given the string ``foo.bar``.

    Calling ``bar = import_item("foo.bar")`` is the functional equivalent of
    executing the code ``from foo import bar``.

    Parameters
    ----------
    name : string
      The fully qualified name of the module/package being imported.

    Returns
    -------
    mod : module object
       The module that was imported.
    """
    
    parts = name.rsplit('.', 1)
    if len(parts) == 2:
        # called with 'foo.bar....'
        package, obj = parts
        module = __import__(package, fromlist=[obj])
        try:
            pak = module.__dict__[obj]
        except KeyError:
            raise ImportError('No module named %s' % obj)
        return pak
    else:
        # called with un-dotted string
        return __import__(parts[0])

if py3compat.PY3:
    def isidentifier(s, dotted=False):
        if dotted:
            return all(isidentifier(a) for a in s.split("."))
        return s.isidentifier()
else:
    _name_re = re.compile(r"[a-zA-Z_][a-zA-Z0-9_]*$")
    def isidentifier(s, dotted=False):
        if dotted:
            return all(isidentifier(a) for a in s.split("."))
        return bool(_name_re.match(s))
    
    
SequenceTypes = (list, tuple, set, frozenset)

#-----------------------------------------------------------------------------
# Basic classes
#-----------------------------------------------------------------------------


class NoDefaultSpecified ( object ): pass
NoDefaultSpecified = NoDefaultSpecified()


class Undefined ( object ): pass
Undefined = Undefined()

class TraitError(Exception):
    pass

#-----------------------------------------------------------------------------
# Utilities
#-----------------------------------------------------------------------------


def class_of ( object ):
    """ Returns a string containing the class name of an object with the
    correct indefinite article ('a' or 'an') preceding it (e.g., 'an Image',
    'a PlotValue').
    """
    if isinstance( object, py3compat.string_types ):
        return add_article( object )

    return add_article( object.__class__.__name__ )


def add_article ( name ):
    """ Returns a string containing the correct indefinite article ('a' or 'an')
    prefixed to the specified string.
    """
    if name[:1].lower() in 'aeiou':
       return 'an ' + name

    return 'a ' + name


def repr_type(obj):
    """ Return a string representation of a value and its type for readable
    error messages.
    """
    the_type = type(obj)
    if (not py3compat.PY3) and the_type is InstanceType:
        # Old-style class.
        the_type = obj.__class__
    msg = '%r %r' % (obj, the_type)
    return msg


def is_trait(t):
    """ Returns whether the given value is an instance or subclass of TraitType.
    """
    return (isinstance(t, TraitType) or
            (isinstance(t, type) and issubclass(t, TraitType)))


def parse_notifier_name(name):
    """Convert the name argument to a list of names.

    Examples
    --------

    >>> parse_notifier_name('a')
    ['a']
    >>> parse_notifier_name(['a','b'])
    ['a', 'b']
    >>> parse_notifier_name(None)
    ['anytrait']
    """
    if isinstance(name, str):
        return [name]
    elif name is None:
        return ['anytrait']
    elif isinstance(name, (list, tuple)):
        for n in name:
            assert isinstance(n, str), "names must be strings"
        return name


class _SimpleTest:
    def __init__ ( self, value ): self.value = value
    def __call__ ( self, test  ):
        return test == self.value
    def __repr__(self):
        return "<SimpleTest(%r)" % self.value
    def __str__(self):
        return self.__repr__()


def getmembers(object, predicate=None):
    """A safe version of inspect.getmembers that handles missing attributes.

    This is useful when there are descriptor based attributes that for
    some reason raise AttributeError even though they exist.  This happens
    in zope.inteface with the __provides__ attribute.
    """
    results = []
    for key in dir(object):
        try:
            value = getattr(object, key)
        except AttributeError:
            pass
        else:
            if not predicate or predicate(value):
                results.append((key, value))
    results.sort()
    return results


#-----------------------------------------------------------------------------
# Base TraitType for all traits
#-----------------------------------------------------------------------------


class TraitType(object):
    """A base class for all trait descriptors.

    Notes
    -----
    Our implementation of traits is based on Python's descriptor
    prototol.  This class is the base class for all such descriptors.  The
    only magic we use is a custom metaclass for the main :class:`HasTraits`
    class that does the following:

    1. Sets the :attr:`name` attribute of every :class:`TraitType`
       instance in the class dict to the name of the attribute.
    2. Sets the :attr:`this_class` attribute of every :class:`TraitType`
       instance in the class dict to the *class* that declared the trait.
       This is used by the :class:`This` trait to allow subclasses to
       accept superclasses for :class:`This` values.
    """


    metadata = {}
    default_value = Undefined
    info_text = 'any value'

    def __init__(self, default_value=NoDefaultSpecified, **metadata):
        """Create a TraitType.
        """
        if default_value is not NoDefaultSpecified:
            self.default_value = default_value

        if len(metadata) > 0:
            if len(self.metadata) > 0:
                self._metadata = self.metadata.copy()
                self._metadata.update(metadata)
            else:
                self._metadata = metadata
        else:
            self._metadata = self.metadata

        self.init()

    def init(self):
        pass

    def get_default_value(self):
        """Create a new instance of the default value."""
        return self.default_value

    def instance_init(self, obj):
        """This is called by :meth:`HasTraits.__new__` to finish init'ing.

        Some stages of initialization must be delayed until the parent
        :class:`HasTraits` instance has been created.  This method is
        called in :meth:`HasTraits.__new__` after the instance has been
        created.

        This method trigger the creation and validation of default values
        and also things like the resolution of str given class names in
        :class:`Type` and :class`Instance`.

        Parameters
        ----------
        obj : :class:`HasTraits` instance
            The parent :class:`HasTraits` instance that has just been
            created.
        """
        self.set_default_value(obj)

    def set_default_value(self, obj):
        """Set the default value on a per instance basis.

        This method is called by :meth:`instance_init` to create and
        validate the default value.  The creation and validation of
        default values must be delayed until the parent :class:`HasTraits`
        class has been instantiated.
        """
        # Check for a deferred initializer defined in the same class as the
        # trait declaration or above.
        mro = type(obj).mro()
        meth_name = '_%s_default' % self.name
        for cls in mro[:mro.index(self.this_class)+1]:
            if meth_name in cls.__dict__:
                break
        else:
            # We didn't find one. Do static initialization.
            dv = self.get_default_value()
            newdv = self._validate(obj, dv)
            obj._trait_values[self.name] = newdv
            return
        # Complete the dynamic initialization.
        obj._trait_dyn_inits[self.name] = cls.__dict__[meth_name]

    def __get__(self, obj, cls=None):
        """Get the value of the trait by self.name for the instance.

        Default values are instantiated when :meth:`HasTraits.__new__`
        is called.  Thus by the time this method gets called either the
        default value or a user defined value (they called :meth:`__set__`)
        is in the :class:`HasTraits` instance.
        """
        if obj is None:
            return self
        else:
            try:
                value = obj._trait_values[self.name]
            except KeyError:
                # Check for a dynamic initializer.
                if self.name in obj._trait_dyn_inits:
                    value = obj._trait_dyn_inits[self.name](obj)
                    # FIXME: Do we really validate here?
                    value = self._validate(obj, value)
                    obj._trait_values[self.name] = value
                    return value
                else:
                    raise TraitError('Unexpected error in TraitType: '
                        'both default value and dynamic initializer are '
                        'absent.')
            except Exception:
                # HasTraits should call set_default_value to populate
                # this.  So this should never be reached.
                raise TraitError('Unexpected error in TraitType: '
                                    'default value not set properly')
            else:
                return value

    def __set__(self, obj, value):
        new_value = self._validate(obj, value)
        old_value = self.__get__(obj)
        obj._trait_values[self.name] = new_value
        if old_value != new_value:
            obj._notify_trait(self.name, old_value, new_value)

    def _validate(self, obj, value):
        if hasattr(self, 'validate'):
            return self.validate(obj, value)
        elif hasattr(self, 'is_valid_for'):
            valid = self.is_valid_for(value)
            if valid:
                return value
            else:
                raise TraitError('invalid value for type: %r' % value)
        elif hasattr(self, 'value_for'):
            return self.value_for(value)
        else:
            return value

    def info(self):
        return self.info_text

    def error(self, obj, value):
        if obj is not None:
            e = "The '%s' trait of %s instance must be %s, but a value of %s was specified." \
                % (self.name, class_of(obj),
                   self.info(), repr_type(value))
        else:
            e = "The '%s' trait must be %s, but a value of %r was specified." \
                % (self.name, self.info(), repr_type(value))
        raise TraitError(e)

    def get_metadata(self, key):
        return getattr(self, '_metadata', {}).get(key, None)

    def set_metadata(self, key, value):
        getattr(self, '_metadata', {})[key] = value


#-----------------------------------------------------------------------------
# The HasTraits implementation
#-----------------------------------------------------------------------------


class MetaHasTraits(type):
    """A metaclass for HasTraits.

    This metaclass makes sure that any TraitType class attributes are
    instantiated and sets their name attribute.
    """

    def __new__(mcls, name, bases, classdict):
        """Create the HasTraits class.

        This instantiates all TraitTypes in the class dict and sets their
        :attr:`name` attribute.
        """
        # print "MetaHasTraitlets (mcls, name): ", mcls, name
        # print "MetaHasTraitlets (bases): ", bases
        # print "MetaHasTraitlets (classdict): ", classdict
        for k,v in py3compat.iteritems(classdict):
            if isinstance(v, TraitType):
                v.name = k
            elif inspect.isclass(v):
                if issubclass(v, TraitType):
                    vinst = v()
                    vinst.name = k
                    classdict[k] = vinst
        return super(MetaHasTraits, mcls).__new__(mcls, name, bases, classdict)

    def __init__(cls, name, bases, classdict):
        """Finish initializing the HasTraits class.

        This sets the :attr:`this_class` attribute of each TraitType in the
        class dict to the newly created class ``cls``.
        """
        for k, v in py3compat.iteritems(classdict):
            if isinstance(v, TraitType):
                v.this_class = cls
        super(MetaHasTraits, cls).__init__(name, bases, classdict)

class HasTraits(py3compat.with_metaclass(MetaHasTraits, object)):

    #__metaclass__ = MetaHasTraits

    def __new__(cls, *args, **kw):
        # This is needed because in Python 2.6 object.__new__ only accepts
        # the cls argument.
        new_meth = super(HasTraits, cls).__new__
        if new_meth is object.__new__:
            inst = new_meth(cls)
        else:
            inst = new_meth(cls, **kw)
        inst._trait_values = {}
        inst._trait_notifiers = {}
        inst._trait_dyn_inits = {}
        # Here we tell all the TraitType instances to set their default
        # values on the instance.
        for key in dir(cls):
            # Some descriptors raise AttributeError like zope.interface's
            # __provides__ attributes even though they exist.  This causes
            # AttributeErrors even though they are listed in dir(cls).
            try:
                value = getattr(cls, key)
            except AttributeError:
                pass
            else:
                if isinstance(value, TraitType):
                    value.instance_init(inst)

        return inst

    def __init__(self, *args, **kw):
        # Allow trait values to be set using keyword arguments.
        # We need to use setattr for this to trigger validation and
        # notifications.
        for key, value in py3compat.iteritems(kw):
            setattr(self, key, value)

    def _notify_trait(self, name, old_value, new_value):

        # First dynamic ones
        callables = self._trait_notifiers.get(name,[])
        more_callables = self._trait_notifiers.get('anytrait',[])
        callables.extend(more_callables)

        # Now static ones
        try:
            cb = getattr(self, '_%s_changed' % name)
        except:
            pass
        else:
            callables.append(cb)

        # Call them all now
        for c in callables:
            # Traits catches and logs errors here.  I allow them to raise
            if callable(c):
                argspec = inspect.getargspec(c)
                nargs = len(argspec[0])
                # Bound methods have an additional 'self' argument
                # I don't know how to treat unbound methods, but they
                # can't really be used for callbacks.
                if isinstance(c, types.MethodType):
                    offset = -1
                else:
                    offset = 0
                if nargs + offset == 0:
                    c()
                elif nargs + offset == 1:
                    c(name)
                elif nargs + offset == 2:
                    c(name, new_value)
                elif nargs + offset == 3:
                    c(name, old_value, new_value)
                else:
                    raise TraitError('a trait changed callback '
                                        'must have 0-3 arguments.')
            else:
                raise TraitError('a trait changed callback '
                                    'must be callable.')


    def _add_notifiers(self, handler, name):
        if name not in self._trait_notifiers:
            nlist = []
            self._trait_notifiers[name] = nlist
        else:
            nlist = self._trait_notifiers[name]
        if handler not in nlist:
            nlist.append(handler)

    def _remove_notifiers(self, handler, name):
        if name in self._trait_notifiers:
            nlist = self._trait_notifiers[name]
            try:
                index = nlist.index(handler)
            except ValueError:
                pass
            else:
                del nlist[index]

    def on_trait_change(self, handler, name=None, remove=False):
        """Setup a handler to be called when a trait changes.

        This is used to setup dynamic notifications of trait changes.

        Static handlers can be created by creating methods on a HasTraits
        subclass with the naming convention '_[traitname]_changed'.  Thus,
        to create static handler for the trait 'a', create the method
        _a_changed(self, name, old, new) (fewer arguments can be used, see
        below).

        Parameters
        ----------
        handler : callable
            A callable that is called when a trait changes.  Its
            signature can be handler(), handler(name), handler(name, new)
            or handler(name, old, new).
        name : list, str, None
            If None, the handler will apply to all traits.  If a list
            of str, handler will apply to all names in the list.  If a
            str, the handler will apply just to that name.
        remove : bool
            If False (the default), then install the handler.  If True
            then unintall it.
        """
        if remove:
            names = parse_notifier_name(name)
            for n in names:
                self._remove_notifiers(handler, n)
        else:
            names = parse_notifier_name(name)
            for n in names:
                self._add_notifiers(handler, n)

    @classmethod
    def class_trait_names(cls, **metadata):
        """Get a list of all the names of this classes traits.

        This method is just like the :meth:`trait_names` method, but is unbound.
        """
        return cls.class_traits(**metadata).keys()

    @classmethod
    def class_traits(cls, **metadata):
        """Get a list of all the traits of this class.

        This method is just like the :meth:`traits` method, but is unbound.

        The TraitTypes returned don't know anything about the values
        that the various HasTrait's instances are holding.

        This follows the same algorithm as traits does and does not allow
        for any simple way of specifying merely that a metadata name
        exists, but has any value.  This is because get_metadata returns
        None if a metadata key doesn't exist.
        """
        traits = dict([memb for memb in getmembers(cls) if \
                     isinstance(memb[1], TraitType)])

        if len(metadata) == 0:
            return traits

        for meta_name, meta_eval in metadata.items():
            if type(meta_eval) is not FunctionType:
                metadata[meta_name] = _SimpleTest(meta_eval)

        result = {}
        for name, trait in traits.items():
            for meta_name, meta_eval in metadata.items():
                if not meta_eval(trait.get_metadata(meta_name)):
                    break
            else:
                result[name] = trait

        return result

    def trait_names(self, **metadata):
        """Get a list of all the names of this classes traits."""
        return self.traits(**metadata).keys()

    def traits(self, **metadata):
        """Get a list of all the traits of this class.

        The TraitTypes returned don't know anything about the values
        that the various HasTrait's instances are holding.

        This follows the same algorithm as traits does and does not allow
        for any simple way of specifying merely that a metadata name
        exists, but has any value.  This is because get_metadata returns
        None if a metadata key doesn't exist.
        """
        traits = dict([memb for memb in getmembers(self.__class__) if \
                     isinstance(memb[1], TraitType)])

        if len(metadata) == 0:
            return traits

        for meta_name, meta_eval in metadata.items():
            if type(meta_eval) is not FunctionType:
                metadata[meta_name] = _SimpleTest(meta_eval)

        result = {}
        for name, trait in traits.items():
            for meta_name, meta_eval in metadata.items():
                if not meta_eval(trait.get_metadata(meta_name)):
                    break
            else:
                result[name] = trait

        return result

    def trait_metadata(self, traitname, key):
        """Get metadata values for trait by key."""
        try:
            trait = getattr(self.__class__, traitname)
        except AttributeError:
            raise TraitError("Class %s does not have a trait named %s" %
                                (self.__class__.__name__, traitname))
        else:
            return trait.get_metadata(key)

#-----------------------------------------------------------------------------
# Actual TraitTypes implementations/subclasses
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# TraitTypes subclasses for handling classes and instances of classes
#-----------------------------------------------------------------------------


class ClassBasedTraitType(TraitType):
    """A trait with error reporting for Type, Instance and This."""

    def error(self, obj, value):
        kind = type(value)
        if (not py3compat.PY3) and kind is InstanceType:
            msg = 'class %s' % value.__class__.__name__
        else:
            msg = '%s (i.e. %s)' % ( str( kind )[1:-1], repr( value ) )

        if obj is not None:
            e = "The '%s' trait of %s instance must be %s, but a value of %s was specified." \
                % (self.name, class_of(obj),
                   self.info(), msg)
        else:
            e = "The '%s' trait must be %s, but a value of %r was specified." \
                % (self.name, self.info(), msg)

        raise TraitError(e)


class Type(ClassBasedTraitType):
    """A trait whose value must be a subclass of a specified class."""

    def __init__ (self, default_value=None, klass=None, allow_none=True, **metadata ):
        """Construct a Type trait

        A Type trait specifies that its values must be subclasses of
        a particular class.

        If only ``default_value`` is given, it is used for the ``klass`` as
        well.

        Parameters
        ----------
        default_value : class, str or None
            The default value must be a subclass of klass.  If an str,
            the str must be a fully specified class name, like 'foo.bar.Bah'.
            The string is resolved into real class, when the parent
            :class:`HasTraits` class is instantiated.
        klass : class, str, None
            Values of this trait must be a subclass of klass.  The klass
            may be specified in a string like: 'foo.bar.MyClass'.
            The string is resolved into real class, when the parent
            :class:`HasTraits` class is instantiated.
        allow_none : boolean
            Indicates whether None is allowed as an assignable value. Even if
            ``False``, the default value may be ``None``.
        """
        if default_value is None:
            if klass is None:
                klass = object
        elif klass is None:
            klass = default_value

        if not (inspect.isclass(klass) or isinstance(klass, py3compat.string_types)):
            raise TraitError("A Type trait must specify a class.")

        self.klass       = klass
        self._allow_none = allow_none

        super(Type, self).__init__(default_value, **metadata)

    def validate(self, obj, value):
        """Validates that the value is a valid object instance."""
        try:
            if issubclass(value, self.klass):
                return value
        except:
            if (value is None) and (self._allow_none):
                return value

        self.error(obj, value)

    def info(self):
        """ Returns a description of the trait."""
        if isinstance(self.klass, py3compat.string_types):
            klass = self.klass
        else:
            klass = self.klass.__name__
        result = 'a subclass of ' + klass
        if self._allow_none:
            return result + ' or None'
        return result

    def instance_init(self, obj):
        self._resolve_classes()
        super(Type, self).instance_init(obj)

    def _resolve_classes(self):
        if isinstance(self.klass, py3compat.string_types):
            self.klass = import_item(self.klass)
        if isinstance(self.default_value, py3compat.string_types):
            self.default_value = import_item(self.default_value)

    def get_default_value(self):
        return self.default_value


class DefaultValueGenerator(object):
    """A class for generating new default value instances."""

    def __init__(self, *args, **kw):
        self.args = args
        self.kw = kw

    def generate(self, klass):
        return klass(*self.args, **self.kw)


class Instance(ClassBasedTraitType):
    """A trait whose value must be an instance of a specified class.

    The value can also be an instance of a subclass of the specified class.
    """

    def __init__(self, klass=None, args=None, kw=None,
                 allow_none=True, **metadata ):
        """Construct an Instance trait.

        This trait allows values that are instances of a particular
        class or its sublclasses.  Our implementation is quite different
        from that of enthough.traits as we don't allow instances to be used
        for klass and we handle the ``args`` and ``kw`` arguments differently.

        Parameters
        ----------
        klass : class, str
            The class that forms the basis for the trait.  Class names
            can also be specified as strings, like 'foo.bar.Bar'.
        args : tuple
            Positional arguments for generating the default value.
        kw : dict
            Keyword arguments for generating the default value.
        allow_none : bool
            Indicates whether None is allowed as a value.

        Default Value
        -------------
        If both ``args`` and ``kw`` are None, then the default value is None.
        If ``args`` is a tuple and ``kw`` is a dict, then the default is
        created as ``klass(*args, **kw)``.  If either ``args`` or ``kw`` is
        not (but not both), None is replace by ``()`` or ``{}``.
        """

        self._allow_none = allow_none

        if (klass is None) or (not (inspect.isclass(klass) or isinstance(klass, py3compat.string_types))):
            raise TraitError('The klass argument must be a class'
                                ' you gave: %r' % klass)
        self.klass = klass

        # self.klass is a class, so handle default_value
        if args is None and kw is None:
            default_value = None
        else:
            if args is None:
                # kw is not None
                args = ()
            elif kw is None:
                # args is not None
                kw = {}

            if not isinstance(kw, dict):
                raise TraitError("The 'kw' argument must be a dict or None.")
            if not isinstance(args, tuple):
                raise TraitError("The 'args' argument must be a tuple or None.")

            default_value = DefaultValueGenerator(*args, **kw)

        super(Instance, self).__init__(default_value, **metadata)

    def validate(self, obj, value):
        if value is None:
            if self._allow_none:
                return value
            self.error(obj, value)

        if isinstance(value, self.klass):
            return value
        else:
            self.error(obj, value)

    def info(self):
        if isinstance(self.klass, py3compat.string_types):
            klass = self.klass
        else:
            klass = self.klass.__name__
        result = class_of(klass)
        if self._allow_none:
            return result + ' or None'

        return result

    def instance_init(self, obj):
        self._resolve_classes()
        super(Instance, self).instance_init(obj)

    def _resolve_classes(self):
        if isinstance(self.klass, py3compat.string_types):
            self.klass = import_item(self.klass)

    def get_default_value(self):
        """Instantiate a default value instance.

        This is called when the containing HasTraits classes'
        :meth:`__new__` method is called to ensure that a unique instance
        is created for each HasTraits instance.
        """
        dv  = self.default_value
        if isinstance(dv, DefaultValueGenerator):
            return dv.generate(self.klass)
        else:
            return dv


class This(ClassBasedTraitType):
    """A trait for instances of the class containing this trait.

    Because how how and when class bodies are executed, the ``This``
    trait can only have a default value of None.  This, and because we
    always validate default values, ``allow_none`` is *always* true.
    """

    info_text = 'an instance of the same type as the receiver or None'

    def __init__(self, **metadata):
        super(This, self).__init__(None, **metadata)

    def validate(self, obj, value):
        # What if value is a superclass of obj.__class__?  This is
        # complicated if it was the superclass that defined the This
        # trait.
        if isinstance(value, self.this_class) or (value is None):
            return value
        else:
            self.error(obj, value)


#-----------------------------------------------------------------------------
# Basic TraitTypes implementations/subclasses
#-----------------------------------------------------------------------------


class Any(TraitType):
    default_value = None
    info_text = 'any value'


class Int(TraitType):
    """An int trait."""

    default_value = 0
    info_text = 'an int'

    def validate(self, obj, value):
        if isinstance(value, int):
            return value
        self.error(obj, value)

class CInt(Int):
    """A casting version of the int trait."""

    def validate(self, obj, value):
        try:
            return int(value)
        except:
            self.error(obj, value)

if py3compat.PY3:
    Long, CLong = Int, CInt
    Integer = Int
else:
    class Long(TraitType):
        """A long integer trait."""

        default_value = long(0)
        info_text = 'a long'

        def validate(self, obj, value):
            if isinstance(value, long):
                return value
            if isinstance(value, int):
                return long(value)
            self.error(obj, value)


    class CLong(Long):
        """A casting version of the long integer trait."""

        def validate(self, obj, value):
            try:
                return long(value)
            except:
                self.error(obj, value)

    class Integer(TraitType):
        """An integer trait.

        Longs that are unnecessary (<= sys.maxint) are cast to ints."""

        default_value = 0
        info_text = 'an integer'

        def validate(self, obj, value):
            if isinstance(value, int):
                return value
            if isinstance(value, long):
                # downcast longs that fit in int:
                # note that int(n > sys.maxint) returns a long, so
                # we don't need a condition on this cast
                return int(value)
            if sys.platform == "cli":
                from System import Int64
                if isinstance(value, Int64):
                    return int(value)
            self.error(obj, value)


class Float(TraitType):
    """A float trait."""

    default_value = 0.0
    info_text = 'a float'

    def validate(self, obj, value):
        if isinstance(value, float):
            return value
        if isinstance(value, int):
            return float(value)
        self.error(obj, value)


class CFloat(Float):
    """A casting version of the float trait."""

    def validate(self, obj, value):
        try:
            return float(value)
        except:
            self.error(obj, value)

class Complex(TraitType):
    """A trait for complex numbers."""

    default_value = 0.0 + 0.0j
    info_text = 'a complex number'

    def validate(self, obj, value):
        if isinstance(value, complex):
            return value
        if isinstance(value, (float, int)):
            return complex(value)
        self.error(obj, value)


class CComplex(Complex):
    """A casting version of the complex number trait."""

    def validate (self, obj, value):
        try:
            return complex(value)
        except:
            self.error(obj, value)

# We should always be explicit about whether we're using bytes or unicode, both
# for Python 3 conversion and for reliable unicode behaviour on Python 2. So
# we don't have a Str type.
class Bytes(TraitType):
    """A trait for byte strings."""

    default_value = b''
    info_text = 'a string'

    def validate(self, obj, value):
        if isinstance(value, bytes):
            return value
        self.error(obj, value)


class CBytes(Bytes):
    """A casting version of the byte string trait."""

    def validate(self, obj, value):
        try:
            return bytes(value)
        except:
            self.error(obj, value)


class Unicode(TraitType):
    """A trait for unicode strings."""

    default_value = py3compat.u('')
    info_text = 'a unicode string'

    def validate(self, obj, value):
        if isinstance(value, py3compat.text_type):
            return value
        if isinstance(value, py3compat.binary_type):
            return unicode(value)
        self.error(obj, value)


class CUnicode(Unicode):
    """A casting version of the unicode trait."""

    def validate(self, obj, value):
        try:
            return unicode(value)
        except:
            self.error(obj, value)


class ObjectName(TraitType):
    """A string holding a valid object name in this version of Python.

    This does not check that the name exists in any scope."""
    info_text = "a valid object identifier in Python"

    if py3compat.PY3:
        # Python 3:
        coerce_str = staticmethod(lambda _,s: s)

    else:
        # Python 2:
        def coerce_str(self, obj, value):
            "In Python 2, coerce ascii-only unicode to str"
            if isinstance(value, unicode):
                try:
                    return str(value)
                except UnicodeEncodeError:
                    self.error(obj, value)
            return value

    def validate(self, obj, value):
        value = self.coerce_str(obj, value)

        if isinstance(value, str) and isidentifier(value):
            return value
        self.error(obj, value)

class DottedObjectName(ObjectName):
    """A string holding a valid dotted object name in Python, such as A.b3._c"""
    def validate(self, obj, value):
        value = self.coerce_str(obj, value)

        if isinstance(value, str) and isidentifier(value, dotted=True):
            return value
        self.error(obj, value)


class Bool(TraitType):
    """A boolean (True, False) trait."""

    default_value = False
    info_text = 'a boolean'

    def validate(self, obj, value):
        if isinstance(value, bool):
            return value
        self.error(obj, value)


class CBool(Bool):
    """A casting version of the boolean trait."""

    def validate(self, obj, value):
        try:
            return bool(value)
        except:
            self.error(obj, value)


class Enum(TraitType):
    """An enum that whose value must be in a given sequence."""

    def __init__(self, values, default_value=None, allow_none=True, **metadata):
        self.values = values
        self._allow_none = allow_none
        super(Enum, self).__init__(default_value, **metadata)

    def validate(self, obj, value):
        if value is None:
            if self._allow_none:
                return value

        if value in self.values:
                return value
        self.error(obj, value)

    def info(self):
        """ Returns a description of the trait."""
        result = 'any of ' + repr(self.values)
        if self._allow_none:
            return result + ' or None'
        return result

class CaselessStrEnum(Enum):
    """An enum of strings that are caseless in validate."""

    def validate(self, obj, value):
        if value is None:
            if self._allow_none:
                return value

        if not isinstance(value, basestring):
            self.error(obj, value)

        for v in self.values:
            if v.lower() == value.lower():
                return v
        self.error(obj, value)

class Container(Instance):
    """An instance of a container (list, set, etc.)

    To be subclassed by overriding klass.
    """
    klass = None
    _valid_defaults = SequenceTypes
    _trait = None

    def __init__(self, trait=None, default_value=None, allow_none=True,
                **metadata):
        """Create a container trait type from a list, set, or tuple.

        The default value is created by doing ``List(default_value)``,
        which creates a copy of the ``default_value``.

        ``trait`` can be specified, which restricts the type of elements
        in the container to that TraitType.

        If only one arg is given and it is not a Trait, it is taken as
        ``default_value``:

        ``c = List([1,2,3])``

        Parameters
        ----------

        trait : TraitType [ optional ]
            the type for restricting the contents of the Container.  If unspecified,
            types are not checked.

        default_value : SequenceType [ optional ]
            The default value for the Trait.  Must be list/tuple/set, and
            will be cast to the container type.

        allow_none : Bool [ default True ]
            Whether to allow the value to be None

        **metadata : any
            further keys for extensions to the Trait (e.g. config)

        """
        # allow List([values]):
        if default_value is None and not is_trait(trait):
            default_value = trait
            trait = None

        if default_value is None:
            args = ()
        elif isinstance(default_value, self._valid_defaults):
            args = (default_value,)
        else:
            raise TypeError('default value of %s was %s' %(self.__class__.__name__, default_value))

        if is_trait(trait):
            self._trait = trait() if isinstance(trait, type) else trait
            self._trait.name = 'element'
        elif trait is not None:
            raise TypeError("`trait` must be a Trait or None, got %s"%repr_type(trait))

        super(Container,self).__init__(klass=self.klass, args=args,
                                  allow_none=allow_none, **metadata)

    def element_error(self, obj, element, validator):
        e = "Element of the '%s' trait of %s instance must be %s, but a value of %s was specified." \
            % (self.name, class_of(obj), validator.info(), repr_type(element))
        raise TraitError(e)

    def validate(self, obj, value):
        value = super(Container, self).validate(obj, value)
        if value is None:
            return value

        value = self.validate_elements(obj, value)

        return value

    def validate_elements(self, obj, value):
        validated = []
        if self._trait is None or isinstance(self._trait, Any):
            return value
        for v in value:
            try:
                v = self._trait.validate(obj, v)
            except TraitError:
                self.element_error(obj, v, self._trait)
            else:
                validated.append(v)
        return self.klass(validated)


class List(Container):
    """An instance of a Python list."""
    klass = list

    def __init__(self, trait=None, default_value=None, minlen=0, maxlen=sys.maxsize,
                allow_none=True, **metadata):
        """Create a List trait type from a list, set, or tuple.

        The default value is created by doing ``List(default_value)``,
        which creates a copy of the ``default_value``.

        ``trait`` can be specified, which restricts the type of elements
        in the container to that TraitType.

        If only one arg is given and it is not a Trait, it is taken as
        ``default_value``:

        ``c = List([1,2,3])``

        Parameters
        ----------

        trait : TraitType [ optional ]
            the type for restricting the contents of the Container.  If unspecified,
            types are not checked.

        default_value : SequenceType [ optional ]
            The default value for the Trait.  Must be list/tuple/set, and
            will be cast to the container type.

        minlen : Int [ default 0 ]
            The minimum length of the input list

        maxlen : Int [ default sys.maxsize ]
            The maximum length of the input list

        allow_none : Bool [ default True ]
            Whether to allow the value to be None

        **metadata : any
            further keys for extensions to the Trait (e.g. config)

        """
        self._minlen = minlen
        self._maxlen = maxlen
        super(List, self).__init__(trait=trait, default_value=default_value,
                                allow_none=allow_none, **metadata)

    def length_error(self, obj, value):
        e = "The '%s' trait of %s instance must be of length %i <= L <= %i, but a value of %s was specified." \
            % (self.name, class_of(obj), self._minlen, self._maxlen, value)
        raise TraitError(e)

    def validate_elements(self, obj, value):
        length = len(value)
        if length < self._minlen or length > self._maxlen:
            self.length_error(obj, value)

        return super(List, self).validate_elements(obj, value)


class Set(Container):
    """An instance of a Python set."""
    klass = set

class Tuple(Container):
    """An instance of a Python tuple."""
    klass = tuple

    def __init__(self, *traits, **metadata):
        """Tuple(*traits, default_value=None, allow_none=True, **medatata)

        Create a tuple from a list, set, or tuple.

        Create a fixed-type tuple with Traits:

        ``t = Tuple(Int, Str, CStr)``

        would be length 3, with Int,Str,CStr for each element.

        If only one arg is given and it is not a Trait, it is taken as
        default_value:

        ``t = Tuple((1,2,3))``

        Otherwise, ``default_value`` *must* be specified by keyword.

        Parameters
        ----------

        *traits : TraitTypes [ optional ]
            the tsype for restricting the contents of the Tuple.  If unspecified,
            types are not checked. If specified, then each positional argument
            corresponds to an element of the tuple.  Tuples defined with traits
            are of fixed length.

        default_value : SequenceType [ optional ]
            The default value for the Tuple.  Must be list/tuple/set, and
            will be cast to a tuple. If `traits` are specified, the
            `default_value` must conform to the shape and type they specify.

        allow_none : Bool [ default True ]
            Whether to allow the value to be None

        **metadata : any
            further keys for extensions to the Trait (e.g. config)

        """
        default_value = metadata.pop('default_value', None)
        allow_none = metadata.pop('allow_none', True)

        # allow Tuple((values,)):
        if len(traits) == 1 and default_value is None and not is_trait(traits[0]):
            default_value = traits[0]
            traits = ()

        if default_value is None:
            args = ()
        elif isinstance(default_value, self._valid_defaults):
            args = (default_value,)
        else:
            raise TypeError('default value of %s was %s' %(self.__class__.__name__, default_value))

        self._traits = []
        for trait in traits:
            t = trait() if isinstance(trait, type) else trait
            t.name = 'element'
            self._traits.append(t)

        if self._traits and default_value is None:
            # don't allow default to be an empty container if length is specified
            args = None
        super(Container,self).__init__(klass=self.klass, args=args,
                                  allow_none=allow_none, **metadata)

    def validate_elements(self, obj, value):
        if not self._traits:
            # nothing to validate
            return value
        if len(value) != len(self._traits):
            e = "The '%s' trait of %s instance requires %i elements, but a value of %s was specified." \
                % (self.name, class_of(obj), len(self._traits), repr_type(value))
            raise TraitError(e)

        validated = []
        for t,v in zip(self._traits, value):
            try:
                v = t.validate(obj, v)
            except TraitError:
                self.element_error(obj, v, t)
            else:
                validated.append(v)
        return tuple(validated)


class Dict(Instance):
    """An instance of a Python dict."""

    def __init__(self, default_value=None, allow_none=True, **metadata):
        """Create a dict trait type from a dict.

        The default value is created by doing ``dict(default_value)``,
        which creates a copy of the ``default_value``.
        """
        if default_value is None:
            args = ((),)
        elif isinstance(default_value, dict):
            args = (default_value,)
        elif isinstance(default_value, SequenceTypes):
            args = (default_value,)
        else:
            raise TypeError('default value of Dict was %s' % default_value)

        super(Dict,self).__init__(klass=dict, args=args,
                                  allow_none=allow_none, **metadata)

class TCPAddress(TraitType):
    """A trait for an (ip, port) tuple.

    This allows for both IPv4 IP addresses as well as hostnames.
    """

    default_value = ('127.0.0.1', 0)
    info_text = 'an (ip, port) tuple'

    def validate(self, obj, value):
        if isinstance(value, tuple):
            if len(value) == 2:
                if isinstance(value[0], py3compat.string_types) and isinstance(value[1], int):
                    port = value[1]
                    if port >= 0 and port <= 65535:
                        return value
        self.error(obj, value)

class CRegExp(TraitType):
    """A casting compiled regular expression trait.

    Accepts both strings and compiled regular expressions. The resulting
    attribute will be a compiled regular expression."""

    info_text = 'a regular expression'

    def validate(self, obj, value):
        try:
            return re.compile(value)
        except:
            self.error(obj, value)



import re
import sys

from ..errors import InvalidMetadataField


_KIND_TO_ABBREVIATED = {
    "cpython": "cp",
    "python": "py",
    "pypy": "pp",
}

_ABBREVIATED_TO_KIND = dict((v, k) for k, v in _KIND_TO_ABBREVIATED.items())

_TAG_RE = re.compile("""
    (?P<interpreter>([^\d]+))
    (?P<version>([\d_]+))
""", flags=re.VERBOSE)


class PythonImplementation(object):
    @classmethod
    def from_running_python(cls):
        s = _abbreviated_implementation() + _implementation_version()
        return cls.from_string(s)

    @classmethod
    def from_string(cls, s):
        generic_exc = InvalidMetadataField('python_tag', s)
        m = _TAG_RE.match(s)
        if m is None:
            raise generic_exc
        else:
            d = m.groupdict()
            kind = d["interpreter"]
            version = d["version"]
            if len(version) < 2:
                raise generic_exc
            elif len(version) > 2:
                raise generic_exc
            else:
                major = int(version[0])
                minor = int(version[1])
                return cls(kind, major, minor)

    def __init__(self, kind, major, minor):
        self.kind = _ABBREVIATED_TO_KIND.get(kind, kind)
        self.major = major
        self.minor = minor

    @property
    def abbreviated_implementation(self):
        return _KIND_TO_ABBREVIATED.get(self.kind, self.kind)

    @property
    def pep425_tag(self):
        """ PEP425-compliant python tag string. """
        return str(self)

    def __str__(self):
        return "{0.abbreviated_implementation}{0.major}{0.minor}".format(self)


def _abbreviated_implementation():
    """Return abbreviated implementation name."""
    if sys.platform.startswith('java'):
        pyimpl = 'jy'
    elif sys.platform == 'cli':
        pyimpl = 'ip'
    elif hasattr(sys, 'pypy_version_info'):
        pyimpl = 'pp'
    else:
        pyimpl = 'cp'
    return pyimpl


def _implementation_version():
    """Return implementation version."""
    return ''.join(map(str, sys.version_info[:2]))



from __future__ import absolute_import

import struct
import sys

from .common import PY2, BytesIO, string_types

if PY2:
    MAX_EXTRACT_VERSION = 63
    from zipfile import (
        BadZipfile as BadZipFile,
        ZipExtFile,
        ZipInfo,
        _CD_COMMENT_LENGTH,
        _CD_EXTRA_FIELD_LENGTH,
        _CD_FILENAME_LENGTH,
        _CD_LOCAL_HEADER_OFFSET,
        _CD_SIGNATURE,
        _ECD_LOCATION,
        _ECD_OFFSET,
        _ECD_SIGNATURE,
        _ECD_SIZE,
        _EndRecData,
        _FH_EXTRA_FIELD_LENGTH,
        _FH_FILENAME_LENGTH,
        _FH_SIGNATURE,
        sizeCentralDir,
        sizeEndCentDir64,
        sizeEndCentDir64Locator,
        sizeFileHeader,
        stringCentralDir,
        stringEndArchive64,
        stringFileHeader,
        structCentralDir,
        structFileHeader,
    )
else:
    from zipfile import (
        BadZipFile,
        MAX_EXTRACT_VERSION,
        ZipExtFile,
        ZipInfo,
        _CD_COMMENT_LENGTH,
        _CD_EXTRA_FIELD_LENGTH,
        _CD_FILENAME_LENGTH,
        _CD_LOCAL_HEADER_OFFSET,
        _CD_SIGNATURE,
        _ECD_LOCATION,
        _ECD_OFFSET,
        _ECD_SIGNATURE,
        _ECD_SIZE,
        _EndRecData,
        _FH_EXTRA_FIELD_LENGTH,
        _FH_FILENAME_LENGTH,
        _FH_SIGNATURE,
        sizeCentralDir,
        sizeEndCentDir64,
        sizeEndCentDir64Locator,
        sizeFileHeader,
        stringCentralDir,
        stringEndArchive64,
        stringFileHeader,
        structCentralDir,
        structFileHeader,
    )

from .common import TooManyFiles

_UTF8_EXTENSION_FLAG = 0x800

if sys.version_info[:2] < (2, 7):
    class _ZipExtFile(ZipExtFile):
        def __enter__(self):
            return self

        def __exit__(self, *a, **kw):
            self.close()


class LeanZipFile(object):
    """ A limited but memory efficient zipfile reader.

    This class has 2 main features:

        * the list of archives is not created explicitly at construction
          time.
        * it will raise an error when iterating over archives with more
          than a given number of members.

    The stdlib ZipFile class creates a list of all the archives in the
    constructor. This may cause issues in situations where you don't want
    to use too much memory, or when you need to handle arbitrary zip files
    (e.g. in a server).
    """
    def __init__(self, filename, max_file_count=None):
        """ Creates a new LeanZipFile instance

        Parameters
        ----------
        filename: str
            Zipfile path
        max_file_count: int or None
            the zipfile has more members than max_file_count, a
            TooManyFiles exception will be raised when accessing members
            beyond this limit. A value of None disable ths limit.
        """
        self.filename = filename
        self.fp = open(filename, 'rb')
        self.max_file_count = max_file_count

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def close(self):
        fp = self.fp
        self.fp = None
        if fp is not None:
            fp.close()

    def get_zip_infos(self, *filenames):
        """Read in the table of contents for the ZIP file."""
        fp = self.fp
        max_file_count = self.max_file_count

        if not fp:
            raise RuntimeError(
                "Attempt to read ZIP archive that was already closed")

        filenames = set(filenames)
        if len(filenames) == 0:
            return

        try:
            endrec = _EndRecData(fp)
        except OSError:
            raise BadZipFile("File is not a zip file")
        if not endrec:
            raise BadZipFile("File is not a zip file")

        size_cd = endrec[_ECD_SIZE]             # bytes in central directory
        offset_cd = endrec[_ECD_OFFSET]         # offset of central directory

        # "concat" is zero, unless zip was concatenated to another file
        concat = endrec[_ECD_LOCATION] - size_cd - offset_cd
        if endrec[_ECD_SIGNATURE] == stringEndArchive64:
            # If Zip64 extension structures are present, account for them
            concat -= (sizeEndCentDir64 + sizeEndCentDir64Locator)

        # start_dir:  Position of start of central directory
        start_dir = offset_cd + concat
        fp.seek(start_dir, 0)
        data = fp.read(size_cd)
        fp = BytesIO(data)
        total = 0
        file_count = 0
        while total < size_cd:
            centdir = fp.read(sizeCentralDir)
            if len(centdir) != sizeCentralDir:
                raise BadZipFile("Truncated central directory")
            centdir = struct.unpack(structCentralDir, centdir)
            if centdir[_CD_SIGNATURE] != stringCentralDir:
                raise BadZipFile("Bad magic number for central directory")
            filename = fp.read(centdir[_CD_FILENAME_LENGTH])
            flags = centdir[5]
            if flags & _UTF8_EXTENSION_FLAG:
                # UTF-8 file names extension
                filename = filename.decode('utf-8')
            else:
                # Historical ZIP filename encoding
                filename = filename.decode('cp437')
            # Create ZipInfo instance to store file information
            x = ZipInfo(filename)
            x.extra = fp.read(centdir[_CD_EXTRA_FIELD_LENGTH])
            x.comment = fp.read(centdir[_CD_COMMENT_LENGTH])
            x.header_offset = centdir[_CD_LOCAL_HEADER_OFFSET]
            (x.create_version, x.create_system, x.extract_version, x.reserved,
             x.flag_bits, x.compress_type, t, d,
             x.CRC, x.compress_size, x.file_size) = centdir[1:12]
            if x.extract_version > MAX_EXTRACT_VERSION:
                raise NotImplementedError("zip file version %.1f" %
                                          (x.extract_version / 10))
            x.volume, x.internal_attr, x.external_attr = centdir[15:18]
            # Convert date/time code to (year, month, day, hour, min, sec)
            x._raw_time = t
            x.date_time = ((d >> 9) + 1980, (d >> 5) & 0xF, d & 0x1F,
                           t >> 11, (t >> 5) & 0x3F, (t & 0x1F) * 2)

            x._decodeExtra()
            x.header_offset = x.header_offset + concat

            # update total bytes read from central directory
            total = (total + sizeCentralDir + centdir[_CD_FILENAME_LENGTH]
                     + centdir[_CD_EXTRA_FIELD_LENGTH]
                     + centdir[_CD_COMMENT_LENGTH])

            file_count += 1
            if max_file_count is not None and file_count > max_file_count:
                raise TooManyFiles('Too many files in egg')

            if x.filename in filenames:
                filenames.discard(x.filename)
                yield x

            if len(filenames) == 0:
                return

    def open(self, zinfo):
        zef_file = self.fp

        if not zef_file:
            raise RuntimeError(
                "Attempt to read ZIP archive that was already closed")

        zef_file.seek(zinfo.header_offset, 0)

        # Skip the file header:
        fheader = zef_file.read(sizeFileHeader)
        if len(fheader) != sizeFileHeader:
            raise BadZipFile("Truncated file header")
        fheader = struct.unpack(structFileHeader, fheader)
        if fheader[_FH_SIGNATURE] != stringFileHeader:
            raise BadZipFile("Bad magic number for file header")

        fname = zef_file.read(fheader[_FH_FILENAME_LENGTH])
        if fheader[_FH_EXTRA_FIELD_LENGTH]:
            zef_file.read(fheader[_FH_EXTRA_FIELD_LENGTH])

        if zinfo.flag_bits & 0x20:
            # Zip 2.7: compressed patched data
            raise NotImplementedError("compressed patched data (flag bit 5)")

        if zinfo.flag_bits & 0x40:
            # strong encryption
            raise NotImplementedError("strong encryption (flag bit 6)")

        if zinfo.flag_bits & _UTF8_EXTENSION_FLAG:
            # UTF-8 filename
            fname_str = fname.decode("utf-8")
        else:
            fname_str = fname.decode("cp437")

        if fname_str != zinfo.orig_filename:
            raise BadZipFile(
                'File name in directory %r and header %r differ.'
                % (zinfo.orig_filename, fname))

        if sys.version_info[:2] < (2, 7):
            return _ZipExtFile(zef_file, zinfo)
        elif sys.version_info[:2] < (3, 4) and sys.platform == 'win32':
                return ZipExtFile(zef_file, 'r', zinfo)
        else:
            return ZipExtFile(zef_file, 'r', zinfo, None, close_fileobj=False)

    def read(self, zinfo_or_name):
        """Return file bytes (as a string) for the given ZipInfo object or
        archive name..

        Parameters
        ----------
        zinfo_or_name: str
            If a string, understood as the full path to the archive name.
            Otherwise, is interpreted as a ZipInfo instance.

        Note
        ----
        Will raise an error if more than one member is found with the
        given name.
        """
        if isinstance(zinfo_or_name, string_types):
            candidates = list(self.get_zip_infos(zinfo_or_name))
            if len(candidates) < 1:
                msg = "There is no item named {0!r} found in the archive"
                raise KeyError(msg.format(zinfo_or_name))
            elif len(candidates) > 1:
                msg = ("There is more than one item named {0!r} found in "
                       "the archive.")
                raise ValueError(msg.format(zinfo_or_name))
            zinfo = candidates[0]
        else:
            zinfo = zinfo_or_name
        with self.open(zinfo) as fp:
            return fp.read()



"""Traitlets-based models for enpkg-related metadata."""
import json
import os

from ..bundled.traitlets import (
    HasTraits, Bool, Enum, Float, Instance, List, Long, Unicode
)
from ..file_formats import EggMetadata, egg_name
from ..file_formats._egg_info import Requirement
from ..file_formats.setuptools_egg import parse_filename
from ..utils import compute_md5
from ..utils.traitlets import NoneOrUnicode


_AVAILABLE_DEFAULT = False
_PRODUCT_DEFAULT = "commercial"


class EnpkgS3IndexEntry(HasTraits):
    """
    Model an S3 legacy index entry.

    Note
    ----
    S3 legacy indexes are the ones generated by our s3 scripts inside
    buildsystem/buildware/scripts/s3, NOT the ones generated by epd_repo.
    """
    available = Bool(_AVAILABLE_DEFAULT)
    build = Long()
    md5 = Unicode()
    mtime = Float()
    egg_basename = Unicode()
    packages = List(Instance(Requirement))
    product = Enum(["pypi", "commercial", "free"], _PRODUCT_DEFAULT)
    python = NoneOrUnicode()
    size = Long()
    type = Enum(["egg"], "egg")
    version = Unicode()

    @classmethod
    def from_data(cls, data):
        """Create a new EnpkgS3IndexEntry instance from a raw dict.

        A raw dict may be a decoded entry from our legacy enpkg S3 index.json

        Note: the passed in dictionary may be modified.
        """
        data["packages"] = [
            Requirement.from_spec_string(s) for s in data.get("packages", [])
        ]
        return cls(**data)

    @classmethod
    def from_egg(cls, path, product=_PRODUCT_DEFAULT,
                 available=_AVAILABLE_DEFAULT):
        kw = {}

        metadata = EggMetadata.from_egg(path)
        kw["version"] = metadata.upstream_version
        kw["build"] = metadata.build
        kw["python"] = metadata._python
        kw["type"] = "egg"
        kw["packages"] = list(metadata.runtime_dependencies)
        kw["product"] = product
        kw["egg_basename"] = metadata.egg_basename

        st = os.stat(path)
        kw["mtime"] = st.st_mtime
        kw["size"] = st.st_size

        kw["available"] = available

        # XXX: keep the hash computing *outside* any other file operation.
        # Opening the same file can cause some IO errors, even on Linux (seen
        # when eggs were on a Samba share).
        kw["md5"] = compute_md5(path)
        return cls(**kw)

    @classmethod
    def from_setuptools_egg(cls, path, build=1, product=_PRODUCT_DEFAULT,
                            available=_AVAILABLE_DEFAULT):
        name, version, pyver, platform = parse_filename(path)
        kw = {"available": True, "build": build, "python": pyver,
              "type": "egg", "version": version}
        st = os.stat(path)
        kw["mtime"] = st.st_mtime
        kw["size"] = st.st_size
        kw["product"] = "free"
        kw["md5"] = compute_md5(path)
        kw["packages"] = []
        kw["egg_basename"] = name

        return cls.from_data(kw)

    @property
    def name(self):
        return self.egg_basename.lower()

    @property
    def s3index_key(self):
        """
        Returns the key in the index.json on S3.
        """
        return egg_name(self.egg_basename, self.version, self.build)

    @property
    def s3index_data(self):
        """
        Returns the data in the index.json on S3.
        """
        data = {
            "available": self.available,
            "build": self.build,
            "md5": self.md5,
            "mtime": self.mtime,
            "name": self.name,
            "packages": [str(p) for p in self.packages],
            "product": self.product,
            "python": self.python,
            "size": self.size,
            "type": self.type,
            "version": self.version,
        }
        return data

    def to_dict(self):
        data = {"available": self.available,
                "build": self.build,
                "egg_basename": self.egg_basename,
                "md5": self.md5,
                "mtime": self.mtime,
                "name": self.name,
                "packages": [str(p) for p in self.packages],
                "product": self.product,
                "python": self.python,
                "size": self.size,
                "type": self.type,
                "version": self.version}
        return data

    def to_json(self):
        return json.dumps(self.to_dict())



from __future__ import absolute_import

import platform
import sys

from ..bundled.traitlets import HasTraits, Enum, Instance, Unicode
from ..errors import OkonomiyakiError

from ._arch import Arch


DARWIN = "darwin"
LINUX = "linux"
SOLARIS = "solaris"
WINDOWS = "windows"

CENTOS = "centos"
DEBIAN = "debian"
RHEL = "rhel"
UBUNTU = "ubuntu"
MAC_OS_X = "mac_os_x"

NAME_TO_PRETTY_NAMES = {
    WINDOWS: "Windows",
    MAC_OS_X: "Mac OS X",
    CENTOS: "CentOS",
    RHEL: "RedHat",
    UBUNTU: "Ubuntu",
    DEBIAN: "Debian",
}

_DIST_NAME_TO_NAME = {
    "centos": CENTOS,
    "redhat": RHEL,
    "ubuntu": UBUNTU,
    "debian": DEBIAN,
}


class Platform(HasTraits):
    """
    An sane generic platform representation.
    """
    os = Enum([WINDOWS, LINUX, DARWIN, SOLARIS])
    """
    The most generic OS description
    """

    name = Enum([WINDOWS, CENTOS, RHEL, DEBIAN, UBUNTU, MAC_OS_X, SOLARIS])
    """
    The most specific platform description
    """

    family = Enum([WINDOWS, RHEL, DEBIAN, MAC_OS_X, SOLARIS])
    """
    The 'kind' of platforms. For example, both debian and ubuntu distributions
    share the same kind, 'debian'.
    """

    release = Unicode()
    """
    The release string. May be empty
    """

    arch = Instance(Arch)
    """
    Actual architecture. The architecture is guessed from the running python.
    """

    machine = Instance(Arch)
    """
    The machine (e.g. 'x86'). This is the CPU architecture (e.g. a 32 bits
    python running on 64 bits Intel OS will be 'amd64', whereas arch will be
    'x86')
    """

    @classmethod
    def from_running_python(cls):
        """ Guess the platform, using the running python to guess the
        architecture.
        """
        return _guess_platform()

    @classmethod
    def from_running_system(cls, arch_string=None):
        """ Guess the platform, with an optional architecture string.

        Parameters
        ----------
        arch_string: str, None
            If given, should be a valid architecture name (e.g. 'x86')
        """
        return _guess_platform(arch_string)

    def __init__(self, os, name, family, arch, machine=None, release=""):
        super(Platform, self).__init__(os=os, name=name, family=family,
                                       arch=arch, machine=machine,
                                       release=release)
        if machine is None:
            self.machine = self.arch

    def __repr__(self):
        return (
            "Platform(os={0.os!r}, name={0.name!r}, family={0.family!r}, "
            "arch='{0.arch}', machine='{0.machine}')".format(self)
        )

    def __str__(self):
        return "{0} {1.release} on {1.machine}".format(
            NAME_TO_PRETTY_NAMES[self.name],
            self
        )

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        else:
            return (self.name == other.name and self.release == other.release
                    and self.arch == other.arch
                    and self.machine == other.machine)

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash((self.name, self.release, self.arch, self.machine))


def _guess_os():
    if sys.platform == "win32":
        return WINDOWS
    elif sys.platform == "darwin":
        return DARWIN
    elif sys.platform.startswith("linux"):
        return LINUX
    else:
        msg = "Could not guess platform from sys.platform: {0!r}"
        raise OkonomiyakiError(msg.format(sys.platform))


def _guess_platform_details(os):
    if os == WINDOWS:
        return WINDOWS, WINDOWS, platform.win32_ver()[0]
    elif os == DARWIN:
        return MAC_OS_X, MAC_OS_X, platform.mac_ver()[0]
    elif os == LINUX:
        name = platform.linux_distribution()[0].lower()
        _, release, _ = platform.dist()
        if name in (DEBIAN, UBUNTU):
            family = DEBIAN
        elif name in (CENTOS, RHEL):
            family = RHEL
        else:
            raise OkonomiyakiError("Unsupported platform: {0!r}".format(name))
        return name, family, release


def _guess_platform(arch_string=None):
    if arch_string is None:
        arch = Arch.from_running_python()
    else:
        arch = Arch.from_name(arch_string)

    machine = Arch.from_running_system()
    os = _guess_os()
    name, family, release = _guess_platform_details(os)

    return Platform(os=os, name=name, family=family, release=release,
                    arch=arch, machine=machine)



from __future__ import absolute_import

from .pkg_info_data import (
    EGG_PKG_INFO_BLACK_LIST, may_be_in_pkg_info_blacklist
)
from .platform_tag import (
    EGG_PLATFORM_BLACK_LIST, may_be_in_platform_blacklist
)
from .python_tag import (
    EGG_PYTHON_TAG_BLACK_LIST, may_be_in_python_tag_blacklist
)


__all__ = [
    "EGG_PLATFORM_BLACK_LIST", "EGG_PKG_INFO_BLACK_LIST",
    "EGG_PYTHON_TAG_BLACK_LIST", "may_be_in_pkg_info_blacklist",
    "may_be_in_python_tag_blacklist", "may_be_in_platform_blacklist",
]



from .pep440 import PEP440Version


class RuntimeVersion(object):
    @classmethod
    def from_string(cls, s):
        return cls(PEP440Version.from_string(s))

    def __init__(self, pep440_version):
        self._pep440_version = pep440_version

    @property
    def major(self):
        return self._nums[0]

    @property
    def minor(self):
        if len(self._nums) < 2:
            return 0
        else:
            return self._nums[1]

    @property
    def micro(self):
        if len(self._nums) < 3:
            return 0
        else:
            return self._nums[2]

    @property
    def numpart(self):
        """Major.Minor.Micro as a string."""
        return "{0.major}.{0.minor}.{0.micro}".format(self)

    @property
    def normalized_string(self):
        return self._pep440_version.normalized_string

    @property
    def _nums(self):
        return self._pep440_version._parts[1]

    def _ensure_can_compare(self, other):
        if not isinstance(other, self.__class__):
            msg = "Cannot compare {0!r} and {1!r}"
            raise TypeError(msg.format(type(self), type(other)))

    def __str__(self):
        return str(self._pep440_version)

    def __hash__(self):
        return hash(self._pep440_version)

    def __eq__(self, other):
        self._ensure_can_compare(other)
        return self._pep440_version == other._pep440_version

    def __ne__(self, other):
        self._ensure_can_compare(other)
        return self._pep440_version != other._pep440_version

    def __lt__(self, other):
        self._ensure_can_compare(other)
        return self._pep440_version < other._pep440_version

    def __le__(self, other):
        self._ensure_can_compare(other)
        return self._pep440_version <= other._pep440_version

    def __ge__(self, other):
        self._ensure_can_compare(other)
        return self._pep440_version >= other._pep440_version

    def __gt__(self, other):
        self._ensure_can_compare(other)
        return self._pep440_version > other._pep440_version



# flake8: noqa
from .abi import default_abi
from .epd_platform import EPDPlatform, applies
from .platform import Platform
from .python_implementation import PythonImplementation
from ._arch import X86, X86_64

__all__ = [
    "EPDPlatform", "Platform", "PythonImplementation", "X86", "X86_64",
    "default_abi"
]



from __future__ import absolute_import

try:
    from ._version import (full_version as __version__,
                           git_revision as __git_revision__,
                           is_released as __is_released__)
except ImportError:
    __version__ = __git_revision__ = "no-built"
    __is_released__ = False

from .common import TooManyFiles
from ._zipfile import (
    PERMS_PRESERVE_NONE, PERMS_PRESERVE_SAFE, PERMS_PRESERVE_ALL, ZipFile
)
from ._lean_zipfile import LeanZipFile
from zipfile import (
    ZIP64_LIMIT, ZIP_DEFLATED, ZIP_FILECOUNT_LIMIT, ZIP_MAX_COMMENT,
    ZIP_STORED,
)

__all__ = [
    "__git_revision__", "__is_released__", "__version__", "LeanZipFile",
    "TooManyFiles", "ZipFile", "PERMS_PRESERVE_NONE", "PERMS_PRESERVE_SAFE",
    "PERMS_PRESERVE_ALL", "ZIP64_LIMIT", "ZIP_DEFLATED", "ZIP_FILECOUNT_LIMIT",
    "ZIP_MAX_COMMENT", "ZIP_STORED",
]



# THIS FILE IS GENERATED FROM ENSTALLER SETUP.PY
version = '4.8.13'
full_version = '4.8.13'
git_revision = '7ba086933768569058f8d800a8e487e7c341f48b'
is_released = True

if not is_released:
    version = full_version



import os
import os.path
import re
import shutil
import tempfile
import time
import zipfile

from egginst.vendor import zipfile2

from ._egg_info import (
    _SPEC_DEPEND_LOCATION, _SPEC_SUMMARY_LOCATION
)
from ._package_info import _PKG_INFO_LOCATION


class _EggBuilderNoPkgInfo(object):
    def __init__(self, egg_metadata, compress=True, cwd=None):
        self.cwd = cwd or os.getcwd()

        if compress is True:
            flag = zipfile.ZIP_DEFLATED
        else:
            flag = zipfile.ZIP_STORED

        self._egg_metadata = egg_metadata
        self._fp = zipfile2.ZipFile(self.path, "w", flag)

        # Write those now so that they are at the beginning of the file.
        self._write_metadata()

    def _write_metadata(self):
        self._write_spec_summary()
        self._write_spec_depend()

    @property
    def path(self):
        return os.path.join(self.cwd, self._egg_metadata.egg_name)

    def close(self):
        self._fp.close()

    def __enter__(self):
        return self

    def __exit__(self, *a, **kw):
        self.commit()

    def add_iterator(self, iterator):
        """
        Add the files specified by the given iterator.

        Parameters
        ----------
        iterator: iterator
            An iterator yielding (path, arcname) pairs.
        """
        for path, arcname in iterator:
            self._fp.write(path, arcname)

    def add_file(self, path, archive_prefix=""):
        """ Add the given file to the egg, under the given archive prefix."""
        arcname = os.path.join(archive_prefix, os.path.basename(path))
        self._fp.write(path, arcname)

    def add_file_as(self, path, archive_name):
        """ Add the given file to the egg, under the given archive name."""
        self._fp.write(path, archive_name)

    def add_data(self, data, archive_name, chmod=0o644):
        """ Write the given data as the given archive name."""
        if archive_name[-1] == "/":
            raise ValueError(
                "Invalid member name for data: '{}'".format(archive_name)
            )

        zinfo = zipfile.ZipInfo(
            filename=archive_name,
            date_time=time.localtime(time.time())[:6]
        )
        zinfo.compress_type = self._fp.compression
        zinfo.external_attr = chmod << 16
        self._fp.writestr(zinfo, data)

    def add_tree(self, directory, archive_prefix=""):
        """
        Add the given directory to the egg, under the given archive_prefix.

        Parameters
        ----------
        directory: path
            A path to a directory. Every file in this directory will be
            included, recursively.
        """
        for root, dirs, files in os.walk(directory):
            for item in dirs + files:
                path = os.path.join(root, item)
                name = os.path.join(archive_prefix,
                                    os.path.relpath(path, directory))
                self._fp.write(path, name)

    def commit(self):
        """ Commit the metadata, and close the file.
        """
        self.close()

    def _write_spec_depend(self):
        spec_depend_string = self._egg_metadata.spec_depend_string
        self.add_data(
            spec_depend_string.encode("ascii"), _SPEC_DEPEND_LOCATION
        )

    def _write_spec_summary(self):
        self.add_data(
            self._egg_metadata.summary.encode("utf8"), _SPEC_SUMMARY_LOCATION
        )

    def _write_pkg_info(self):
        data = self._egg_metadata.pkg_info.to_string()
        self.add_data(data.encode("utf8"), _PKG_INFO_LOCATION)


class EggBuilder(_EggBuilderNoPkgInfo):
    """
    Class to build eggs from an install tree. This is mostly useful to
    build Enthought eggs for non-python packages (C/C++ libraries, etc...)
    """
    def __init__(self, egg_metadata, compress=True, cwd=None):
        if egg_metadata.pkg_info is None:
            msg = ("EggBuilder does not accept EggMetadata instances with "
                   "a None pkg_info attribute.")
            raise ValueError(msg)

        super(EggBuilder, self).__init__(egg_metadata, compress, cwd)

    def _write_metadata(self):
        super(EggBuilder, self)._write_metadata()
        self._write_pkg_info()


def _no_rename(f):
    return f


def _accept_anything(f, nameset):
    return True


class DefaultAcceptFilter(object):
    """ A simple filter that excludes a <name>.py[c|o] file if a <name>.so or
    <name>.pyd exists.
    """
    def __init__(self, funcs=None):
        self._r_compiled = re.compile(r'(.+)\.py(c|o)?$')
        self._funcs = funcs or tuple()

    def _filter_py(self, f, nameset):
        m = self._r_compiled.match(f)
        if m:
            so = m.group(1) + ".so"
            pyd = m.group(1) + ".pyd"
            if so in nameset or pyd in nameset:
                return False
            else:
                return True
        else:
            return True

    def __call__(self, arcname, nameset):
        acc = True
        for f in (self._filter_py,) + self._funcs:
            acc = acc and f(arcname, nameset)
        return acc


class EggRewriter(_EggBuilderNoPkgInfo):
    """ Class to create Enthought eggs from existing setuptools eggs.
    """
    def __init__(self, egg_metadata, egg, compress=True, cwd=None,
                 rename=None, accept=None, allow_overwrite=False):
        """ Create a new egg rewriter instance.

        Parameters
        ----------
        egg_metadata: EggMetadata
            The metadata to use to write Enthought metadata
        egg: str
            Path to the egg to start from. The path must be accessible for
            read.
        compress: bool
            Whether to compress the zipfile
        cwd: path
            The directory where to write the generated egg. If not
            specified, defaults to os.getcwd()
        rename: callable
            If defined, a callable of the form (archive_name, ) ->
            new_archive_name, to rename archive members from the original
            egg.
        accept: callable
            If defined, a callable of the form (archive_name, namelist) ->
            bool, returning True for archives to copy from the original
            egg. Namelist is a set of all the archives in the existing egg. By
            default, uses a DefaultFilter instance.
        allow_overwrite: bool
            By default, the egg creation will fail if one adds existing
            archives. If set to True, one can overwrite archive members
            already present in the source egg.

        Note
        ----
        When both rename and accept arguments are used, the filtre applies
        on the archive name *before* the renaming, i.e. on the archive
        name in the original egg.
        """
        super(EggRewriter, self).__init__(egg_metadata, compress, cwd)
        self._egg = egg
        self._rename = rename or _no_rename
        self._accept = accept or DefaultAcceptFilter()

        self._allow_overwrite = allow_overwrite

    def commit(self):
        self._copy_existing_content()
        super(EggRewriter, self).commit()

    def _copy_existing_content(self):
        with zipfile2.ZipFile(self._egg) as source:
            tempdir = tempfile.mkdtemp()
            try:
                nameset = set(source.namelist())
                for f in source.namelist():
                    arcname = self._rename(f)

                    if self._allow_overwrite:
                        if arcname in self._fp._filenames_set:
                            continue

                    if self._accept(f, nameset):
                        source_path = source.extract(f, tempdir)
                        self.add_file_as(source_path, arcname)
            finally:
                shutil.rmtree(tempdir)



import sys


PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3


if PY2:
    def exec_(_code_, _globs_=None, _locs_=None):
        """Execute code in a namespace."""
        if _globs_ is None:
            frame = sys._getframe(1)
            _globs_ = frame.f_globals
            if _locs_ is None:
                _locs_ = frame.f_locals
            del frame
        elif _locs_ is None:
            _locs_ = _globs_
        exec("""exec _code_ in _globs_, _locs_""")

    long = long
    string_types = basestring,

    binary_type = str
    text_type = unicode

    def u(s):
        return unicode(s.replace(r'\\', r'\\\\'), "unicode_escape")

    def iterkeys(d, **kw):
        return iter(d.iterkeys(**kw))

    def itervalues(d, **kw):
        return iter(d.itervalues(**kw))

    def iteritems(d, **kw):
        return iter(d.iteritems(**kw))

    import StringIO
    StringIO = BytesIO = StringIO.StringIO

else:
    import builtins
    exec_ = getattr(builtins, "exec")

    long = int
    string_types = str,

    binary_type = bytes
    text_type = str

    def u(s):
        return s

    def iterkeys(d, **kw):
        return iter(d.keys(**kw))

    def itervalues(d, **kw):
        return iter(d.values(**kw))

    def iteritems(d, **kw):
        return iter(d.items(**kw))

    import io
    StringIO = io.StringIO


def with_metaclass(meta, *bases):
    """Create a base class with a metaclass."""
    # This requires a bit of explanation: the basic idea is to make a dummy
    # metaclass for one level of class instantiation that replaces itself with
    # the actual metaclass.
    class metaclass(meta):
        def __new__(cls, name, this_bases, d):
            return meta(name, bases, d)
    return type.__new__(metaclass, 'temporary_class', (), {})



from egginst.vendor.okonomiyaki.versions.pep386_workaround import (
    PEP386WorkaroundVersion
)


def normalize_version_string(version_string):
    """
    Normalize the given version string to a string that can be converted to
    a NormalizedVersion.

    This function applies various special cases needed for EPD/Canopy and not
    handled in NormalizedVersion parser.

    Parameters
    ----------
    version_string: str
        The version to convert

    Returns
    -------
    normalized_version: str
        The normalized version string. Note that this is not guaranteed to be
        convertible to a NormalizedVersion
    """
    if version_string.endswith(".dev"):
        version_string += "1"
    return version_string



from __future__ import absolute_import

import abc

import six

from attr import asdict, attr, attributes
from attr.validators import instance_of

from ..errors import UnsupportedMetadata
from ..platforms import Platform
from ..utils import substitute_variable, substitute_variables
from ..versions import MetadataVersion, RuntimeVersion

from .common import _platform_string
from .runtime_metadata import JuliaRuntimeMetadataV1, PythonRuntimeMetadataV1


@attributes
class IRuntimeInfo(six.with_metaclass(abc.ABCMeta)):
    """ The metadata of a runtime package (i.e. the actual zipfile containing
    the runtime code).
    """
    metadata_version = attr(validator=instance_of(MetadataVersion))

    @classmethod
    def factory_from_metadata(cls, metadata, prefix, name):
        """ Creates a runtime info object of the appropriate class.

        The class is selected dynamically from the given metadata.
        """
        return runtime_info_from_metadata(metadata, prefix, name)

    @classmethod
    def factory_from_json_dict(cls, json_dict):
        """ Creates a runtime info object of the appropriate class.

        The class is selected dynamically from the given json dict.
        """
        return runtime_info_from_json(json_dict)

    @abc.abstractmethod
    def to_json_dict(self):
        """ Returns a json dict that can easily be serialized back into a
        file or buffer.
        """


@attributes
class IRuntimeInfoV1(IRuntimeInfo):
    # Note: the attributes in IRuntimeInfoV1 need to be synchronized with
    # IRuntimeMetadataV1
    implementation = attr(validator=instance_of(six.text_type))
    "The implementation (e.g. 'cpython')"

    version = attr(validator=instance_of(RuntimeVersion))
    """The implementation version, e.g. pypy 2.6.1 would report 2.6.1 as the
    'upstream' part."""

    language_version = attr(validator=instance_of(RuntimeVersion))
    """This is the 'language' version, e.g.  pypy 2.6.1 would report 2.7.10
    here."""

    platform = attr(validator=instance_of(Platform))
    "The platform on which this runtime may run."

    abi = attr(validator=instance_of(six.text_type))
    "The ABI of this runtime."

    build_revision = attr(validator=instance_of(six.text_type))
    """The internal version. Informative only, has no semantices and my be
    empty."""

    executable = attr(validator=instance_of(six.text_type))
    """Executable path. May be a templated variable.
    """

    paths = attr(validator=instance_of(tuple))
    """Directories to add to $PATH to have access to this runtime."""

    post_install = attr(validator=instance_of(tuple))
    """Command to execute as part of post installation."""

    prefix = attr(validator=instance_of(six.text_type))
    """ Full path to the installed prefix."""

    name = attr(validator=instance_of(six.text_type))

    _metadata_klass = None

    @classmethod
    def _from_json_dict(cls, data):
        metadata = cls._metadata_klass._from_json_dict(data)
        return cls._from_metadata(
            metadata, data["prefix"], data["name"]
        )

    @classmethod
    def _from_metadata(cls, metadata, prefix, name):
        return cls(*cls._from_metadata_impl(metadata, prefix, name))

    @classmethod
    def _from_metadata_impl(cls, metadata, prefix, name):
        implementation = metadata.implementation
        version = metadata.version
        language_version = metadata.language_version
        platform = metadata.platform
        abi = metadata.abi
        build_revision = metadata.build_revision

        variables = _compute_variables(metadata, prefix, name)

        executable = substitute_variable(metadata.executable, variables)

        paths = tuple(
            substitute_variable(path, variables) for path in metadata.paths
        )
        post_install = tuple(
            substitute_variable(part, variables)
            for part in metadata.post_install
        )

        return (
            MetadataVersion.from_string("1.0"), implementation,
            version, language_version, platform, abi, build_revision,
            executable, paths, post_install, prefix, name
        )

    def to_json_dict(self):
        data_dict = asdict(self)
        for k in ("version", "metadata_version", "language_version"):
            data_dict[k] = str(data_dict[k])
        data_dict["platform"] = _platform_string(self.platform)

        return data_dict


class JuliaRuntimeInfoV1(IRuntimeInfoV1):
    """ Class representing the metadata of an installed julia runtime.
    """
    _metadata_klass = JuliaRuntimeMetadataV1


@attributes
class PythonRuntimeInfoV1(IRuntimeInfoV1):
    """ Class representing the metadata of an installed python runtime.
    """
    scriptsdir = attr(validator=instance_of(six.text_type))
    site_packages = attr(validator=instance_of(six.text_type))
    python_tag = attr(validator=instance_of(six.text_type))

    _metadata_klass = PythonRuntimeMetadataV1

    @classmethod
    def _from_metadata_impl(cls, metadata, prefix, name):
        args = super(PythonRuntimeInfoV1, cls)._from_metadata_impl(
            metadata, prefix, name
        )
        variables = _compute_variables(metadata, prefix, name)
        scriptsdir = substitute_variable(metadata.scriptsdir, variables)
        site_packages = substitute_variable(metadata.site_packages, variables)
        python_tag = substitute_variable(metadata.python_tag, variables)

        return args + (scriptsdir, site_packages, python_tag)


_RUNTIME_INFO_JSON_FACTORY = {
    (MetadataVersion.from_string("1.0"), "pypy"): PythonRuntimeInfoV1,
    (MetadataVersion.from_string("1.0"), "cpython"): PythonRuntimeInfoV1,
    (MetadataVersion.from_string("1.0"), "julia"): JuliaRuntimeInfoV1,
}


def runtime_info_from_json(json_data):
    """ Create a runtime info instance from its json serialized state.
    """
    metadata_version = MetadataVersion.from_string(
        json_data["metadata_version"]
    )
    implementation = json_data["implementation"]
    key = (metadata_version, implementation)
    klass = _RUNTIME_INFO_JSON_FACTORY.get(key)
    if klass is None:
        msg = "Combination {0!r} is not supported".format(key)
        raise UnsupportedMetadata(metadata_version, msg)
    else:
        return klass._from_json_dict(json_data)


_RUNTIME_INFO_FACTORY = {}
for klass in (JuliaRuntimeInfoV1, PythonRuntimeInfoV1):
    _RUNTIME_INFO_FACTORY[klass._metadata_klass] = klass


def runtime_info_from_metadata(metadata, prefix, name):
    """ Create a runtime info instance from a runtime metadata instance, an
    (install) prefix and a runtime name.
    """
    klass = _RUNTIME_INFO_FACTORY.get(type(metadata))
    assert klass is not None
    return klass._from_metadata(metadata, prefix, name)


def _compute_variables(metadata, prefix, name):
    data = dict(
        (k.name, getattr(metadata, k.name)) for k in metadata.__attrs_attrs__
    )

    variables = dict(
        (k, v) for k, v in data.items() if isinstance(v, six.string_types)
    )
    variables["prefix"] = prefix
    variables["name"] = name

    return substitute_variables(variables, variables)



from __future__ import absolute_import

import platform
import sys

from ..bundled.traitlets import HasTraits, Enum
from ..errors import OkonomiyakiError


X86 = "x86"
X86_64 = "x86_64"

_ARCH_NAME_TO_BITS = {
    X86: 32,
    X86_64: 64,
}

_ARCH_NAME_TO_NORMALIZED = {
    "amd64": X86_64,
    "AMD64": X86_64,
    "x86_64": X86_64,
    "x86": X86,
    "i386": X86,
    "i686": X86,
}


class Arch(HasTraits):
    name = Enum([X86, X86_64])
    """
    Actual architecture name (e.g. 'x86'). The architecture is guessed from the
    running python.
    """

    @classmethod
    def _from_bitwidth(cls, bitwidth):
        if bitwidth == "32":
            return cls.from_name(X86)
        elif bitwidth == "64":
            return cls.from_name(X86_64)
        else:
            msg = "Invalid bits width: {0!r}".format(bitwidth)
            raise OkonomiyakiError(msg)

    @classmethod
    def from_name(cls, name):
        # FIXME: kept for backward compatibility
        return cls(name)

    @classmethod
    def from_running_python(cls):
        machine = platform.machine()

        if machine in _ARCH_NAME_TO_NORMALIZED:
            if sys.maxsize > 2 ** 32:
                return Arch.from_name(X86_64)
            else:
                return Arch.from_name(X86)
        else:
            raise OkonomiyakiError("Unknown machine combination {0!r}".
                                   format(machine))

    @classmethod
    def from_running_system(cls):
        return Arch.from_name(platform.machine())

    def __init__(self, name):
        normalized_name = _ARCH_NAME_TO_NORMALIZED.get(name)
        if normalized_name is None:
            msg = "Unsupported/unrecognized architecture: {0!r}"
            raise OkonomiyakiError(msg.format(name))

        super(Arch, self).__init__(name=normalized_name)

    @property
    def bits(self):
        """
        Actual architecture bits (e.g. 32), as an int.
        """
        return _ARCH_NAME_TO_BITS[self.name]

    @property
    def _arch_bits(self):
        # Used by EPDPlatform
        return str(self.bits)

    @property
    def _legacy_name(self):
        # Used for translating the arch into EGG-INFO/spec/depend (old
        # 'platform' entry)
        if self.name == X86_64:
            return "amd64"
        else:
            return self.name

    def __repr__(self):
        return "Arch({0.name!r})".format(self)

    def __str__(self):
        return self.name

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        else:
            return self.name == other.name and self.bits == other.bits

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash((self.name, self.bits))



try:
    from _version import version as __version__
except ImportError:
    __version__ = "no-built"

from ._implementation import HasTraits, MetaHasTraits, TraitType, Any, CBytes, Dict, \
    Int, Long, Integer, Float, Complex, Bytes, Unicode, TraitError, \
    Undefined, Type, This, Instance, TCPAddress, List, Tuple, \
    ObjectName, DottedObjectName, CRegExp, Enum, Bool, CBool



from __future__ import absolute_import, division, print_function

import copy

from ._compat import iteritems
from ._make import Attribute, NOTHING, fields


def asdict(inst, recurse=True, filter=None):
    """
    Return the ``attrs`` attribute values of *i* as a dict.  Optionally recurse
    into other ``attrs``-decorated classes.

    :param inst: Instance of a ``attrs``-decorated class.

    :param recurse: Recurse into classes that are also ``attrs``-decorated.
    :type recurse: bool

    :param filter: A callable whose return code deteremines whether an
        attribute or element is included (``True``) or dropped (``False``).  Is
        called with the :class:`attr.Attribute` as the first argument and the
        value as the second argument.
    :type filer: callable

    :rtype: :class:`dict`
    """
    attrs = fields(inst.__class__)
    rv = {}
    for a in attrs:
        v = getattr(inst, a.name)
        if filter is not None and not filter(a, v):
            continue
        if recurse is True:
            if has(v.__class__):
                rv[a.name] = asdict(v, recurse=True, filter=filter)
            elif isinstance(v, (tuple, list, set)):
                rv[a.name] = [
                    asdict(i, recurse=True, filter=filter)
                    if has(i.__class__) else i
                    for i in v
                ]
            elif isinstance(v, dict):
                rv[a.name] = dict((asdict(kk) if has(kk.__class__) else kk,
                                   asdict(vv) if has(vv.__class__) else vv)
                                  for kk, vv in iteritems(v))
            else:
                rv[a.name] = v
        else:
            rv[a.name] = v
    return rv


def has(cl):
    """
    Check whether *cl* is a class with ``attrs`` attributes.

    :param cl: Class to introspect.
    :type cl: type

    :raise TypeError: If *cl* is not a class.

    :rtype: :class:`bool`
    """
    try:
        fields(cl)
    except ValueError:
        return False
    else:
        return True


def assoc(inst, **changes):
    """
    Copy *inst* and apply *changes*.

    :param inst: Instance of a class with ``attrs`` attributes.

    :param changes: Keyword changes in the new copy.

    :return: A copy of inst with *changes* incorporated.
    """
    new = copy.copy(inst)
    for k, v in iteritems(changes):
        a = getattr(new.__class__, k, NOTHING)
        if a is NOTHING or not isinstance(a, Attribute):
            raise ValueError(
                "{k} is not an attrs attribute on {cl}."
                .format(k=k, cl=new.__class__)
            )
        setattr(new, k, v)
    return new



import os.path
import re
import sys
try:
    import sysconfig
except ImportError:  # Python 2.6 support
    sysconfig = None
import warnings

from ..errors import OkonomiyakiError
from ..platforms import PythonImplementation
from ..utils import py3compat
from ._egg_info import _guess_python_tag
from ._package_info import PackageInfo


_R_EGG_NAME = re.compile("""
        (?P<name>^[^.-]+)
        (-(?P<version>[^-]+))
        (-py(?P<pyver>(\d+\.\d+)))
        (-(?P<platform>.+))?
        \.egg$
""", re.VERBOSE)


def parse_filename(path):
    """
    Parse a setuptools egg.

    Returns
    -------
    name : str
        the egg name
    version : str
        the egg version
    python_version : str
        the python version
    platform : str or None
        the platform string, or None for platform-independent eggs.
    """
    m = _R_EGG_NAME.search(os.path.basename(path))
    if m:
        platform = m.group("platform")
        return (m.group("name"), m.group("version"), m.group("pyver"),
                platform)
    else:
        raise OkonomiyakiError("Invalid egg name: {0}".format(path))


def _get_default_setuptools_abi(platform_string, pyver):
    """ Try to guess the ABI for setuptools eggs from the platform_string
    and pyver parts.

    Parameters
    ----------
    platform_string: str
        The platform part of the setuptools egg filename as a string. If
        None, understood as a cross platform egg (pure python).
    pyver: str
        The python version
    """
    assert pyver in ("2.6", "2.7")

    if platform_string is None:
        return None
    else:
        if (platform_string.startswith("linux")
                or platform_string.startswith("win")
                or platform_string.startswith("macosx")):
            return 'cp{0}{1}m'.format(pyver[0], pyver[2])
        else:
            msg = "Platform string {0!r} not supported".format(platform_string)
            raise ValueError(msg)


_UNSPECIFIED = object()


def _guess_abi_from_python(python):
    # For legacy (aka legacy spec version info < 1.3), we know that pyver
    # can only be one of "2.X" with X in (5, 6, 7).
    #
    # In those cases, the mapping (platform pyver) -> ABI is unambiguous,
    # as we only ever used one ABI for a given python version/platform.
    return "cp{0}{1}m".format(python.major, python.minor)


def _guess_abi_from_running_python():
    if sysconfig is None:
        soabi = None
    else:
        try:
            soabi = sysconfig.get_config_var('SOABI')
        except IOError as e:  # pip issue #1074
            warnings.warn("{0}".format(e), RuntimeWarning)
            soabi = None

    if soabi and soabi.startswith('cpython-'):
        return 'cp' + soabi.split('-', 1)[-1]
    else:
        return None


def _guess_abi(platform, python):
    if platform is None:
        return None

    if python is not None:
        if (python.major, python.minor) != sys.version_info[:2]:
            return _guess_abi_from_python(python)

    abi = _guess_abi_from_running_python()
    if abi is None:
        if python is not None:
            return _guess_abi_from_python(python)
        else:
            msg = ("Could not guess ABI, you need to specify the abi_tag "
                   "argument to from_egg, e.g. 'cp34m' for Enthought "
                   "CPython 3.4 runtimes")
            raise OkonomiyakiError(msg)
    else:
        return abi


class SetuptoolsEggMetadata(object):
    @classmethod
    def from_egg(cls, path, platform=None, python=_UNSPECIFIED,
                 abi_tag=_UNSPECIFIED):
        filename = os.path.basename(path)
        name, version, pyver, platform_string = parse_filename(filename)

        if platform is None and platform_string is not None:
            msg = ("Platform-specific egg detected: you need to specify a "
                   "platform argument that is not None to from_egg")
            raise OkonomiyakiError(msg)

        if python is _UNSPECIFIED:
            python = PythonImplementation.from_string(_guess_python_tag(pyver))

        if abi_tag is _UNSPECIFIED:
            abi_tag = _guess_abi(platform, python)
        else:
            abi_tag = abi_tag

        pkg_info = PackageInfo.from_egg(path)

        return cls(name, version, platform, python, abi_tag, pkg_info)

    def __init__(self, name, version, platform, python, abi_tag, pkg_info):
        """
        Parameters
        ----------
        name: str
            Package name
        version: str
            Version string
        platform: EPDPlatform
            An EPDPlatform instance, or None for cross-platform eggs
        python: PythonImplementation
            The PEP425 python tag, or None.
        abi_tag: str
            The PEP425 abi tag, or None.
        pkg_info: PackageInfo
            Instance representing the egg PKG-INFO.
        """
        self.name = name
        self.version = version

        self.platform = platform
        if isinstance(python, py3compat.string_types):
            python = PythonImplementation.from_string(python)
        self.python = python
        self.abi_tag = abi_tag

        self._pkg_info = pkg_info

    @property
    def platform_tag(self):
        """ Platform tag following PEP425, except that no platform is
        represented as None and not 'any'."""
        if self.platform is None:
            return None
        else:
            return self.platform.pep425_tag

    @property
    def python_tag(self):
        if self.python is None:
            return None
        else:
            return self.python.pep425_tag

    @property
    def summary(self):
        return self._pkg_info.summary



from __future__ import absolute_import, division, print_function

import sys


PY3 = sys.version_info[0] == 3
PY2 = sys.version_info[0] == 2


if PY2:
    # TYPE is used in exceptions, repr(int) is different on Python 2 and 3.
    TYPE = "type"

    def exec_(code, locals_, globals_):
        exec("exec code in locals_, globals_")

    def iteritems(d):
        return d.iteritems()
else:
    TYPE = "class"

    def exec_(code, locals_, globals_):
        exec(code, locals_, globals_)

    def iteritems(d):
        return d.items()



import ast
import contextlib
import shutil
import string
import tempfile

from .py3compat import string_types

from ..errors import OkonomiyakiError


class _AssignmentParser(ast.NodeVisitor):
    def __init__(self):
        self._data = {}

    def parse(self, s):
        self._data.clear()

        root = ast.parse(s)
        self.visit(root)
        return self._data

    def generic_visit(self, node):
        if type(node) != ast.Module:
            raise OkonomiyakiError("Unexpected expression @ line {0}".
                                   format(node.lineno))
        super(_AssignmentParser, self).generic_visit(node)

    def visit_Assign(self, node):
        try:
            value = ast.literal_eval(node.value)
        except ValueError:
            msg = ("Invalid configuration syntax at line {0}".
                   format(node.lineno))
            raise OkonomiyakiError(msg)
        else:
            for target in node.targets:
                self._data[target.id] = value


def parse_assignments(file_or_filename):
    """
    Parse files which contain only python assignements, and returns the
    corresponding dictionary name: value

    Parameters
    ----------
    file_or_filename: str, file object
        If a string, interpreted as a filename. File object otherwise.
    """
    if isinstance(file_or_filename, string_types):
        with open(file_or_filename) as fp:
            return _AssignmentParser().parse(fp.read())
    else:
        return _AssignmentParser().parse(file_or_filename.read())


@contextlib.contextmanager
def tempdir():
    d = tempfile.mkdtemp()
    try:
        yield d
    finally:
        shutil.rmtree(d)


def substitute_variables(d, local_vars):
    """Perform shell/Perl-style variable substitution.

    Every occurrence of '${name}' name is considered a variable, and variable
    is substituted by the value found in the `local_vars' dictionary.  Raise
    ValueError for any variables not found in `local_vars'.

    '$' may be escaped by using '$$'

    Parameters
    ----------
    d: dict
        (str: str) mapping, where each value will be substituted.
    local_vars: dict
        dict of variables
    """
    def _resolve(d):
        ret = {}
        for k, v in d.items():
            ret[k] = substitute_variable(v, local_vars)
        return ret

    ret = _resolve(d)
    while not ret == d:
        d = ret
        ret = _resolve(d)
    return ret


def substitute_variable(v, local_vars):
    return string.Template(v).substitute(local_vars)



# flake8: noqa
_JULIA_V1 = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "title": "PythonRuntimeMetadata v1.0",
    "description": "PythonRuntimeMetadata runtime/metadata.json schema.",
    "type": "object",
    "properties": {
        "metadata_version": {
            "description": "The metadata version.",
            "type": "string"
        },
        "implementation": {
            "description": "The implementation (e.g. cpython)",
            "type": "string"
        },
        "version": {
            "description": "The implementation version, e.g.  pypy 2.6.1 would report 2.6.1 as the 'upstream' part.",
            "type": "string"
        },
        "abi": {
            "description": "The runtime's ABI, e.g. 'msvc2008' or 'gnu'.",
            "type": "string"
        },
        "language_version": {
            "description": "This is the 'language' version, e.g.  pypy 2.6.1 would report 2.7.10 here.",
            "type": "string"
        },
        "platform": {
            "description": ("The platform string (as can be parsed by"
                            "EPDPlatform.from_epd_string"),
            "type": "string"
        },
        "build_revision": {
            "description": "Build revision (internal only).",
            "type": "string",
        },
        "executable": {
            "description": "The full path to the actual runtime executable.",
            "type": "string",
        },
        "paths": {
            "description": "The list of path to have access to this runtime.",
            "type": "array",
            "items": {"type": "string"},
        },
        "post_install": {
            "description": ("The command (as a list) to execute after "
                            "installation."),
            "type": "array",
            "items": {"type": "string"},
        },
    },
    "required": [
        "metadata_version",
        "implementation",
        "version",
        "abi",
        "language_version",
        "platform",
        "build_revision",
        "executable",
        "paths",
        "post_install",
    ]
}

_PYTHON_V1 = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "title": "PythonRuntimeMetadata v1.0",
    "description": "PythonRuntimeMetadata runtime/metadata.json schema.",
    "type": "object",
    "properties": {
        "metadata_version": {
            "description": "The metadata version.",
            "type": "string"
        },
        "implementation": {
            "description": "The implementation (e.g. cpython)",
            "type": "string"
        },
        "version": {
            "description": "The implementation version, e.g.  pypy 2.6.1 would report 2.6.1 as the 'upstream' part.",
            "type": "string"
        },
        "abi": {
            "description": "The runtime's ABI, e.g. 'msvc2008' or 'gnu'.",
            "type": "string"
        },
        "language_version": {
            "description": "This is the 'language' version, e.g.  pypy 2.6.1 would report 2.7.10 here.",
            "type": "string"
        },
        "platform": {
            "description": ("The platform string (as can be parsed by"
                            "EPDPlatform.from_epd_string"),
            "type": "string"
        },
        "build_revision": {
            "description": "Build revision (internal only).",
            "type": "string",
        },
        "executable": {
            "description": "The full path to the actual runtime executable.",
            "type": "string",
        },
        "paths": {
            "description": "The list of path to have access to this runtime.",
            "type": "array",
            "items": {"type": "string"},
        },
        "post_install": {
            "description": ("The command (as a list) to execute after "
                            "installation."),
            "type": "array",
            "items": {"type": "string"},
        },
        "scriptsdir": {
            "description": "Full path to scripts directory.",
            "type": "string",
        },
        "site_packages": {
            "description": "The full path to the python site packages.",
            "type": "string",
        },
        "python_tag": {
            "description": "The python tag, as defined in PEP 425.",
            "type": "string",
        },
    },
    "required": [
        "metadata_version",
        "implementation",
        "version",
        "abi",
        "language_version",
        "platform",
        "build_revision",
        "executable",
        "paths",
        "post_install",
        "scriptsdir",
        "site_packages",
        "python_tag",
    ]
}



import ntpath
import posixpath
import sys

import six

from attr import attr, attributes
from attr.validators import instance_of

from ..platforms import Platform, default_abi
from ..platforms.platform import WINDOWS
from ..versions import MetadataVersion, RuntimeVersion

from .runtime_info import IRuntimeInfoV1, PythonRuntimeInfoV1


# Those *Runtime classes are not very useful at the moment, and don't really
# belong in okonomiyaki. We need them here as they are required for enstaller,
# but will be removed once jaguar and enstaller are merged together.
@attributes
class Runtime(object):
    """ Runtime instances are used to manipulate installed runtimes.
    """
    _runtime_info = attr(validator=instance_of(IRuntimeInfoV1))

    @property
    def executable(self):
        "The full path to the runtime interpreter."
        return self._metadata_klass.executable

    @property
    def prefix(self):
        "The full path prefix to this runtime."
        return self._runtime_info.prefix


@attributes
class PythonRuntime(Runtime):
    """ A runtime with python-specific attributes.

    It also implements some functionality for enstaller legacy support.
    """
    _executable = attr(validator=instance_of(six.text_type))

    @classmethod
    def from_prefix_and_platform(cls, prefix, platform, version=None,
                                 python_tag=None):
        """ Use this to build a runtime for an arbitrary platform.

        Calling this with an incompatible platform (e.g. windows on linux) is
        undefined.

        Parameters
        ----------
        prefix: text
            An absolute path to the prefix (the root of the runtime), e.g. for
            a standard unix python, if python is in <prefix>/bin/python,
            <prefix> is the prefix.
        platform: Platform
            An okonomiyaki Platform class (the vendorized one).
        implementation_version: RuntimeVersion
            The runtime's implementation version. Default to a version
            representing sys.version_info if not specified
        """
        version = version or _version_info_to_version()
        language_version = RuntimeVersion.from_string(version.numpart)

        if hasattr(sys, 'pypy_version_info'):
            implementation = "pypy"
            python_tag_prefix = "pp"
            implementation_version = RuntimeVersion.from_string(
                "{0}+1".format(
                    ".".join(
                        str(i) for i in sys.pypy_version_info[:3]
                    )
                )
            )
        else:
            implementation = "cpython"
            python_tag_prefix = "cp"
            implementation_version = version

        python_tag = (
            python_tag or u"{0}{1}{2}".format(
                python_tag_prefix, version.major, version.minor
            )
        )
        abi = default_abi(platform, implementation, implementation_version)

        if six.PY2:
            prefix = prefix.decode(sys.getfilesystemencoding())

        if platform.os == WINDOWS:
            prefix = ntpath.normpath(prefix)
            scriptsdir = ntpath.join(prefix, "Scripts")
            paths = (prefix, scriptsdir)
        else:
            prefix = posixpath.normpath(prefix)
            scriptsdir = bindir = posixpath.join(prefix, "bin")
            paths = (bindir, )

        if version.major == 3:
            executable = u"python3"
        else:
            executable = u"python"

        major_minor = "{0}.{1}".format(version.major, version.minor)

        if platform.os == WINDOWS:
            executable += ".exe"
            executable = ntpath.join(prefix, executable)
        else:
            executable = posixpath.join(scriptsdir, executable)

        site_packages = _compute_site_packages(prefix, platform, major_minor)

        implementation = u"cpython"
        build_revision = u""
        name = u"<dummy>"
        post_install = tuple()

        runtime_info = PythonRuntimeInfoV1(
            MetadataVersion.from_string("1.0"), implementation,
            version, language_version, platform, abi, build_revision,
            executable, paths, post_install, prefix, name, scriptsdir,
            site_packages, python_tag,

        )
        return cls(runtime_info, u"")

    @classmethod
    def from_running_python(cls, platform=None):
        """ Use this to compute runtime info from the running python.

        Calling this with an incompatible platform (e.g. windows on linux) is
        undefined.

        Parameters
        ----------
        platform: Platform
            An okonomiyaki Platform class (the vendorized one).
        """
        platform = platform or Platform.from_running_python()
        version = _version_info_to_version()

        return cls.from_prefix_and_platform(sys.exec_prefix, platform, version)

    @property
    def executable(self):
        "The full path to the python binary."
        if len(self._executable) == 0:
            self._executable = self._compute_executable()
        return self._executable

    @property
    def scriptsdir(self):
        """The full path to the scripts directory, i.e. the directory where
        setuptools entry points are usually placed.
        """
        return self._runtime_info.scriptsdir

    @property
    def site_packages(self):
        """The full path to this runtime site-packages.
        """
        return self._runtime_info.site_packages

    def _compute_executable(self):
        if self._runtime_info.platform.os == WINDOWS:
            # Hack to take into account virtualenvs
            paths = (
                self._runtime_info.executable,
                ntpath.join(
                    self.scriptsdir,
                    ntpath.basename(self._runtime_info.executable)
                )
            )
            for path in paths:
                if ntpath.isfile(path):
                    return path
            return self._runtime_info.executable
        else:
            return self._runtime_info.executable


def _compute_site_packages(prefix, platform, major_minor):
    # Adapted from distutils.sysconfig.get_python_lib for 2.7.9
    prefix = prefix or sys.exec_prefix

    if platform.os == WINDOWS:
        return ntpath.join(prefix, "Lib", "site-packages")
    else:
        return posixpath.join(
            prefix, "lib", "python" + major_minor, "site-packages"
        )


def _version_info_to_version(version_info=None):
    version_info = version_info or sys.version_info
    version_string = ".".join(str(part) for part in version_info[:3])
    version_string += "+{0}.{1}".format(*version_info[-2:])
    return RuntimeVersion.from_string(version_string)
